#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulation pipeline runner** (i.e., simulation activity iteratively
run by its parent pipeline) functionality.
'''

# ....................{ IMPORTS                            }....................
from betse.exceptions import BetseSimPipeException
from betse.science.phase.require.abc.phasereqset import (
    SimPhaseRequirements, SimPhaseRequirementsOrNoneTypes)
from betse.util.type.decorator.deccls import MethodDecoratorABC
from betse.util.type.text import strs
from betse.util.type.types import (
    type_check, CallableTypes, SequenceTypes,)

# ....................{ CLASSES                            }....................
class SimPipeRunner(MethodDecoratorABC):
    '''
    Class decorator annotating simulation pipeline runners with custom metadata.

    All such runners decorated by the :func:`piperunner` decorator are
    guaranteed to be instances of this class, which provides all metadata passed
    to that decorator as instance variables of the same name.

    Attributes
    ----------
    categories : SequenceTypes
        Sequence of one or more human-readable strings iteratively naming all
        arbitrary categories to which this runner belongs (in descending order
        of hierarchical taxonomy).
    method_name : str
        Name of the method implementing this runner.
    requirements : SimPhaseRequirements
        Immutable set of zero or more :class:`SimPhaseRequirement` instances
        specifying all simulation features required by this runner.
    description : str
        Human-readable description of this runner as a **single-line string**
        (i.e., containing no newlines).

    See Also
    ----------
    :func:`piperunner`
        Further details.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        method: CallableTypes,
        categories: SequenceTypes,
        requirements: SimPhaseRequirementsOrNoneTypes,
    ) -> None:
        '''
        Initialize this class decorator.

        Parameters
        ----------
        method: CallableTypes
            Unbound method (i.e., function) to be decorated.
        categories : SequenceTypes
            Sequence of one or more human-readable category names.
        requirements : SimPhaseRequirementsOrNoneTypes
            Immutable set of zero or more :class:`SimPhaseRequirement` instances
            *or* ``None``, in which case this parameter defaults to the empty
            immutable set of such instances.

        Raises
        ----------
        BetseSimPipeException
            If this method has no docstring.
        '''

        # Initialize our superclass with the passed method.
        super().__init__(method)

        # Default all unpassed parameters to sane defaults.
        if requirements is None:
            requirements = SimPhaseRequirements()

        # Classify all passed parameters.
        self.requirements = requirements

        # Classify all remaining passed parameters.
        self.categories = categories
        self.method_name = method.__name__

        # Default this runner's description to its docstring.
        self.description = method.__doc__

        # If this docstring is empty, raise an exception.
        if not self.description:
            raise BetseSimPipeException(
                'Runner method {}() has no docstring.'.format(method.__name__))
        # Else, this docstring is non-empty.

        # Transform this docstring into a description by...
        self.description = (
            # Removing all leading and trailing whitespace.
            strs.remove_whitespace_presuffix(
            # Reducing from a (possibly) multi- to single-line string.
            strs.unwrap(self.description)))

    # ..................{ CALLERS                            }..................
    @type_check
    def __call__(
        self,

        # To avoid circular import dependencies, this is type-checked as a
        # fully-qualified class name resolved at runtime.
        pipeline: 'betse.science.phase.pipe.pipeabc.SimPipeABC',
        *args,
        **kwargs
    ) -> object:

        # If this runner is unsatisfied, raise an exception.
        pipeline.die_unless_runner_satisfied(self)
        # Else, this runner is satisfied. Since the prior call already logged
        # the attempt to run this runner, avoid redoing so here.

        # Defer to the superclass implementation to run this runner.
        return super().__call__(pipeline, *args, **kwargs)

# ....................{ DECORATORS                         }....................
@type_check
def piperunner(
    # Mandatory metadata.
    categories: SequenceTypes,

    # Optional metadata.
    requirements: SimPhaseRequirementsOrNoneTypes = None,
) -> CallableTypes:
    '''
    Decorator annotating simulation pipeline **runners** (i.e., methods of
    :class:`SimPipeABC` subclasses with names prefixed by
    :attr:`SimPipeABC._RUNNER_METHOD_NAME_PREFIX`) with custom metadata.

    All methods decorated by this decorator are guaranteed to be instances of
    the :class:`SimPipeRunner` class, which provides all metadata passed to
    this decorator as instance variables of the same name.

    Caveats
    ----------
    **This decorator is strictly optional.** Runners *not* decorated by this
    decorator are still runnable from simulation pipelines. Since this decorator
    annotates runners with metadata, however, unannotated runners will *not* be
    usable by external interfaces expecting this metadata -- typically, GUIs
    populating interactive widget fields by this metadata.

    **Runner methods decorated by this decorator should not be decorated by
    other decorators.** In particular, decorated methods should *not* also be
    decorated by :func:`@type_check`, which this decorator already internally
    decorates all decorated methods by.

    Parameters
    ----------
    categories : SequenceTypes
        Sequence of one or more human-readable strings iteratively naming all
        arbitrary categories to which this runner belongs (in descending order
        of hierarchical taxonomy). Categories are arbitrary labels accessed
        *only* by external interfaces and are otherwise ignored by the core
        codebase. Specifically:
        * The first string in this sequence names an arbitrary **root category**
          (e.g., root node in a tree view), intended to be shared between
          multiple runners. This string is typically a broadly applicable label
          such as ``Voltage Plots``.
        * The last string in this sequence names an arbitrary **leaf category**
          (e.g., leaf node in a tree view), intended to be unique to a single
          runner. This string is typically a narrowly applicable label such as
          ``Extracellular Voltage Plot``.
        * All other strings in this sequence name arbitrary categories of
          increasingly fine-grained depth, again intended to be shared between
          multiple runners.
    requirements : SimPhaseRequirementsOrNoneTypes
        Immutable set of zero or more :class:`SimPhaseRequirement` instances
        specifying all simulation features required by this runner. This
        decorator then decorates this runner by performing the following logic
        immediately *before* calling this runner:
        * For each requirement in this set...
          * If this requirement is unsatisfied by the current simulation phase
            (e.g., as the configuration for this phase disables extracellular
            spaces), this requirement (and hence this runner) is unsatisfied.
            Since this constitutes a fatal error, an
            :class:`BetseSimPipeRunnerUnsatisfiedException` is raised.
          * Else, this runner is run.
        Defaults to ``None``, in which case no such decoration is applied.
    '''

    @type_check
    def _piperunner_closure(method: CallableTypes) -> SimPipeRunner:
        '''
        Closure both type-checking *and* annotating the passed simulation
        pipeline runner method with the metadata passed to the outer decorator
        defining this closure, returning an instance of the class decorator
        exposing this metadata to external interfaces.

        Parameters
        ----------
        method : CallableTypes
            Unbound method (i.e., function) to be decorated by (in order):
            #. The :func:`@type_check` decorator, type checking this method.
               For efficiency, callers should ensure this method is *not*
               externally decorated by this decorator.
            #. The :class:`SimPipeRunner` class decorator,
               annotating this method with this metadata.

        See Also
        ----------
        :func:`piperunner`
            Further details.
        '''

        # Return an instance of the class decorator exposing this metadata.
        return SimPipeRunner(
            # As a caller convenience, ensure this method is type-checked.
            method=type_check(method),
            categories=categories,
            requirements=requirements,
        )

    # Return the closure accepting the method to be decorated.
    return _piperunner_closure



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **mapping utilities** (i.e., functions operating on dictionary-like
types and instances).
'''

# ....................{ IMPORTS                           }....................
import pprint
from betse.exceptions import BetseMappingException
from betse.util.type.types import (
    type_check, MappingType, HashableType,)
from copy import deepcopy

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_values_unique(mapping: MappingType) -> None:
    '''
    Raise an exception unless all values of the passed dictionary are unique.

    Equivalently, this function raises an exception if any two key-value pairs
    of this dictionary share the same values.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inspected.

    Raises
    ----------
    BetseMappingException
        If at least one value of this dictionary is a duplicate.
    '''

    # Avoid circular import dependencies.
    from betse.util.type import iterables
    from betse.util.type.text import strs

    # If one or more values of this dictionary are duplicates...
    if not is_values_unique(mapping):
        # Set of all duplicate values in this dictionary.
        values_duplicate = iterables.get_items_duplicate(mapping.values())

        # Raise an exception embedding this set.
        raise BetseMappingException(
            'Dictionary values {} duplicate.'.format(
                strs.join_as_conjunction_double_quoted(*values_duplicate)))

# ....................{ TESTERS                           }....................
@type_check
def is_key(mapping: MappingType, *keys: HashableType) -> bool:
    '''
    ``True`` only if the passed dictionary contains *all* passed keys.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be tested.
    keys : tuple[HashableType]
        Tuple of all keys to be tested for.

    Returns
    ----------
    bool
        ``True`` only if this dictionary contains *all* passed keys.
    '''

    return (
        # If only one key is passed, optimize this common edge case with the
        # standard idiom for testing key containment.
        keys[0] in mapping if len(keys) == 1 else
        # Else, two or more keys are passed. In this case, fallback to a
        # general-case strategy testing key containment in a single one-liner.
        # And yes: this is ridiculously awesome.
        set(keys).issubset(mapping)
    )


@type_check
def is_values_unique(mapping: MappingType) -> bool:
    '''
    ``True`` only if all values of the passed dictionary are **unique** (i.e.,
    if *no* two key-value pairs of this dictionary share the same values).

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inspected.

    Returns
    ----------
    bool
        ``True`` only if *all* values of this dictionary are unique.
    '''

    # Avoid circular import dependencies.
    from betse.util.type import iterables

    # For sanity, defer to an existing low-level tester.
    return iterables.is_items_unique(mapping.values())

# ....................{ FORMATTERS                        }....................
@type_check
def format_map(mapping: MappingType) -> str:
    '''
    Convert the passed dictionary into a human-readable string.
    '''

    return pprint.pformat(mapping)

# ....................{ COPIERS                           }....................
@type_check
def copy_map(mapping: MappingType) -> MappingType:
    '''
    Dictionary of all key-value pairs deeply (i.e., recursively) duplicated
    from the passed dictionary.

    This function should *always* be called in lieu of the standard
    :meth:`dict.__init__` and :meth:`dict.copy` methods, which only perform
    shallow dictionary copies. These copies fail to copy data structures nested
    in the values of the original dictionary, inviting subtle synchronization
    woes on subsequently modifying either the original or copied dictionaries.

    Parameters
    ----------
    mapping: MappingType
        Dictionary to be deeply copied.

    Returns
    ----------
    MappingType
        Dictionary of all key-value pairs deeply (i.e., recursively) duplicated
        from the passed dictionary.
    '''

    #FIXME: Does this simplistic approach guarantee the returned mapping to be
    #of the same type as the passed mapping?
    return deepcopy(mapping)


@type_check
def copy_map_sans_key(mapping: MappingType, key: HashableType) -> MappingType:
    '''
    Dictionary of all key-value pairs excluding that whose key is the passed
    key deeply (i.e., recursively) duplicated from the passed dictionary.

    Parameters
    ----------
    mapping: MappingType
        Dictionary to be deeply copied.
    key : HashableType
        Key to be removed from this dictionary.

    Returns
    ----------
    MappingType
        Dictionary of all key-value pairs excluding that whose key is this
        key deeply (i.e., recursively) duplicated from this dictionary.

    Raises
    ----------
    :class:`KeyError`
        If this dictionary contains no such key.

    See Also
    ----------
    :func:`copy_map`
        Further details on map copying.
    :func:`remove_key`
        Further details on key removal.
    '''

    # Deep copy of this dictionary.
    mapping_copy = copy_map(mapping)

    # Remove this key from this copy in-place.
    remove_key(mapping=mapping_copy, key=key)

    # Return this copy.
    return mapping_copy

# ....................{ INVERTERS                         }....................
@type_check
def invert_map_unique(mapping: MappingType) -> MappingType:
    '''
    Dictionary inverted from the passed dictionary if no two key-value pairs of
    this dictionary share the same values *or* raise an exception otherwise.

    Specifically, the returned dictionary maps from each value to each key of
    the passed dictionary *and* is guaranteed to be the same type as that of
    the passed dictionary.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inverted. The type of this dictionary *must* define an
        ``__init__`` method accepting a single parameter whose value is an
        iterable of 2-iterables ``(key, value)`` providing all key-value pairs
        with which to initialize a new such dictionary. See the
        :meth:`dict.__init__` method for further details.

    Returns
    ----------
    MappingType
        Dictionary inverted from this dictionary as detailed above.

    Raises
    ----------
    BetseMappingException
        If one or more key-value pairs of this dictionary share the same
        values.

    See Also
    ----------
    https://stackoverflow.com/a/1679702/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # If any values of this dictionary are are duplicates, raise an exception.
    die_unless_values_unique(mapping)

    # Type of this dictionary.
    mapping_type = type(mapping)

    # If this is an unordered dictionary, return a dictionary comprehension
    # efficiently inverting this dictionary in the trivial way.
    if mapping_type is dict:
        return {value: key for key, value in mapping.items()}
    # Else, this is possibly an ordered dictionary. In this case, a
    # considerably less trivial and slightly less efficient approach is needed.
    else:
        # Iterable of reversed 2-iterables "(value, pair)" for each key-value
        # pair of the passed dictionary. Dismantled, this is:
        #
        # * "mapping.items()", an iterable of 2-iterables "(key, value)" for
        #   each key-value pair of the passed dictionary.
        # * "reversed", a builtin which when passed such a 2-iterable returns
        #   the reversed 2-iterable "(value, pair)" for that key-value pair.
        # * "map(...)", a builtin applying the prior builtin to each such pair.
        value_key_pairs = map(reversed, mapping.items())

        # Return a new instance of this type of dictionary by invoking the
        # "dict(iterable)" form of this type's __init__() method. To quote the
        # dict.__init__() docstring:
        #
        # "dict(iterable) -> new dictionary initialized as if via:
        #      d = {}
        #      for k, v in iterable:
        #          d[k] = v"
        return mapping_type(value_key_pairs)

# ....................{ MERGERS                           }....................
@type_check
def merge_maps(*mappings: MappingType) -> MappingType:
    '''
    Dictionary of all key-value pairs deeply (i.e., recursively) merged
    together from all passed dictionaries (in the passed order).

    **Order is significant.** Dictionaries passed later take precedence over
    dictionaries passed earlier. Ergo, the last passed dictionary takes
    precedence over *all* other passed dictionaries. Whenever any two passed
    dictionaries collide (i.e., contain the same key), the returned dictionary
    contains a key-value pair for that key whose value is that of the key-value
    pair for the same key of whichever of the two dictionaries was passed last.

    Parameters
    ----------
    mappings : Tuple[MappingType]
        Tuple of all dictionaries to be merged.

    Returns
    ----------
    MappingType
        Dictionary merged from and of the same type as the passed dictionaries.
        Note lastly that the class of the passed dictionary *must* define an
        ``__init__()`` method accepting a dictionary comprehension.

    See Also
    ----------
    :meth:`dict.update`
        Standard method merging two dictionaries, which should typically be
        called instead of this slower function in this specific use case.
    '''

    # Type of dictionary to be returned.
    dict_type = type(mappings[0])

    # Dictionary merged from the passed dictionaries via a doubly-nested
    # dictionary comprehension. While there exist a countably infinite number
    # of approaches to merging dictionaries in Python, this approach is known
    # to be the most efficient for general-purpose merging of arbitrarily many
    # dictionaries under Python >= 3.4. See also Trey Hunter's exhaustive
    # commentary replete with timings at:
    #     http://treyhunner.com/2016/02/how-to-merge-dictionaries-in-python
    dict_merged = {
        # For safety, deeply copy rather than reuse this value.
        key: deepcopy(value)
        for mapping in mappings
        for key, value in mapping.items()
    }

    # Return a dictionary of this type converted from this dictionary. If the
    # desired type is a "dict", this dictionary is returned as is; else, this
    # dictionary is converted into an instance of the desired type.
    return dict_merged if dict_type is dict else dict_type(dict_merged)

# ....................{ REMOVERS                          }....................
@type_check
def remove_key(mapping: MappingType, key: HashableType) -> None:
    '''
    Remove the key-value pair whose key is the passed key from the passed
    dictionary **in-place** (i.e., by modifying this dictionary rather than
    creating and returning a new dictionary with this key removed) if this
    dictionary contains this key *or* raise an exception otherwise.

    This function is a caller convenience improving codebase readability and
    efficiency. Although there exist multiple means of removing key-value pairs
    from dictionaries, this function implements the most efficient approach.
    These include:

    * The ``del mapping[key]`` idiom, known to be the most efficient approach.
    * The :meth:`dict.pop` method, known to be slightly less efficient than the
      idiomatic approach.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to remove this key from.
    key : HashableType
        Key to be removed from this dictionary.

    Raises
    ----------
    :class:`KeyError`
        If this dictionary contains no such key.

    See Also
    ----------
    :func:`copy_map_sans_key`
        Function creating and returning a new dictionary with this key removed.
    '''

    # The best things in life are free.
    del mapping[key]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level support facilities for Numpy, a mandatory runtime dependency.
'''

# ....................{ IMPORTS                            }....................
from betse.util.io.log import logs
from betse.util.os import dlls, oses
from betse.util.path import dirs, files, pathnames
from betse.util.type import iterables, modules
from betse.util.type.mapping.mapcls import OrderedArgsDict
from betse.util.type.text import regexes, strs
from betse.util.type.types import BoolOrNoneTypes
from numpy import __config__ as numpy_config

# ....................{ GLOBALS ~ opt_info                 }....................
# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX = None
'''
Uncompiled regular expression heuristically matching the basenames of optimized
BLAS shared libraries in the `libraries` list of the global
:data:`numpy.__config__.blas_opt_info` dictionary.

This expression does _not_ match the strict superset of optimized BLAS shared
libraries that are also optimized, as doing so in a cross-platform manner is
infeasible. Debian-based Linux distributions, for example, remove the prefixes
uniquely identifying the threaded variants of both ATLAS and OpenBLAS from the
basenames of their shared libraries (e.g., "libatlas.so" rather than
"libtatlas.so").

Optimized BLAS shared libraries matched by this expression include:

* **AMD Core Math Library (ACML), unconditionally GPU- and CPU- optimized in
  both OpenMP-based and non-OpenMP-based variants regardless of underlying
  compiler (e.g., GNU Fortran, Open64). Note, however, that ACML does _not_ ship
  with a CBLAS interface and hence is non-trivial to link Numpy against. While
  unlikely that any end users will ever do so, it nonetheless remains feasible
  and hence somewhat supported.
* **Automatically Tuned Linear Algebra Software (ATLAS),** both single- and
  multithreaded CBLAS- and Fortran-based variants for both ATLAS < 3.10 and
  ATLAS >= 3.10, which ships shared libraries under different basenames than
  ATLAS < 3.10. (Life complicates life.)
* **Intel Math Kernel Library (MKL),** unconditionally multithreaded in both
  OpenMP-based and non-OpenMP-based variants regardless of underlying compiler
  (e.g., dynamic, GCC, Intel).  Note that **Intel Vector Mathematical Library
  (VML)** is intentionally ignored.  Although also unconditionally
  multithreaded, VML does _not_ implement the BLAS API. Numpy currently contains
  no VML-specific handling, apart from (somewhat uselessly) detecting VML
  installation on reporting system diagnostics.
* **OpenBLAS,** both single- and multithreaded 32- and 64-bit variants.

This expression is typically only required once at application startup and hence
is conditionally compiled in a just-in-time (JIT) manner by the
:func:`_is_blas_optimized_python_general` function rather than
unconditionally compiled here.
'''
# print('blas regex: ' + _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX)


# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX = None
'''
Uncompiled regular expression heuristically matching the dirnames of optimized
BLAS shared libraries in the `libraries` list of the global
:data:`numpy.__config__.blas_opt_info` dictionary.

See Also
----------
:data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX`
    Further details.
'''


_OPTIMIZED_BLAS_OPT_INFO_EXTRA_LINK_ARGS_MACOS = {
    # Accelerate. Although Accelerate is only conditionally multithreaded,
    # multithreading is enabled by default and hence a safe assumption.
    '-Wl,Accelerate',

    # vecLib. Similar assumptions as with Accelerate apply.
    '-Wl,vecLib',
}
'''
Set of all strings in the `extra_link_args` list of the global
:data:`numpy.__config__.blas_opt_info` dictionary heuristically corresponding to
macOS-specific optimized BLAS shared libraries.

Unlike all other such libraries, Numpy does _not_ declare unique dictionary
globals describing macOS-specific BLAS shared libraries when linked against.
Hence, this lower-level solution.
'''

# ....................{ GLOBALS ~ linked lib               }....................
# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX = None
'''
Uncompiled regular expression heuristically matching the basenames of optimized
BLAS shared libraries dynamically linked to by Numpy.

See Also
----------
:data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX`
    Further details.
'''


# Fully initialized by the _init_globals() function below.
_OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX = None
'''
Uncompiled regular expression heuristically matching the dirnames of optimized
BLAS shared libraries dynamically linked to by Numpy.

See Also
----------
:data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX`
    Further details.
'''

# ....................{ INITIALIZERS                       }....................
# For simplicity, this function is called below on the first importation of this
# submodule rather than explicitly called by callers.
def init() -> None:
    '''
    Initialize this submodule.

    Specifically (_in order_):

    . Initialize all uninitialized global variables of this submodule.
    . If the currently installed version of Numpy was linked against an
      unoptimized BLAS implementation and is thus itself unoptimized, log a
      non-fatal warning.
    '''

    # Initialize all uninitialized global variables of this submodule.
    _init_globals()

    # If Numpy linked against an unoptimized BLAS, log a non-fatal warning.
    if not is_blas_optimized():
        logs.log_warning(
            'Numpy unoptimized; scaling down to single-core operation. '
            'Consider installing an optimized multithreaded '
            'CBLAS implementation (e.g., OpenBLAS, ATLAS, ACML, MKL) and '
            'reinstalling Numpy to use this implementation.'
        )


def _init_globals() -> None:
    '''
    Initialize all uninitialized global variables of this submodule.
    '''

    # Permit these globals to be redefined.
    global\
        _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX,\
        _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX,\
        _OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX,\
        _OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX

    # Regular expression fragment matching uniquely identifying substrings of
    # basenames of optimized BLAS shared libraries.
    blas_lib_basename_regex = r'|'.join((
        # AMD Core Math Library (ACML).
        r'acml',

        # Automatically Tuned Linear Algebra Software (ATLAS) >= 3.10.
        r'[st]?atlas',

        # Automatically Tuned Linear Algebra Software (ATLAS) < 3.10.
        r'(?:pt)?f77blas',

        # Automatically Tuned Linear Algebra Software (ATLAS) < 3.10.  Although
        # some platforms (e.g., Ubuntu) distribute the CBLAS implementation of
        # ATLAS with the ambiguous basename of "cblas" rather than "ptcblas",
        # the former may also refer to the unoptimized reference CBLAS
        # implementation and is hence ignored.
        r'ptcblas',

        # Intel Math Kernel Library (MKL). Thanks to the profusion of possible
        # library basenames, this regular expression fragment simplistically
        # assumes *ALL* library basenames prefixed by "mkl" to unconditionally
        # connote MKL. What could go wrong?
        r'mkl',

        # OpenBLAS.
        r'openblas',

        #FIXME: Research. No idea if this basename substring is even accurate.

        # BLAS-like Library Instantiation Software (BLIS). Unconditionally
        # multithreaded. Technically, Numpy has yet to add official support for
        # BLIS. Since numerous contributors nonetheless perceive BLIS to be the
        # eventual successor of BLAS *AND* since Numpy currently hosts an open
        # pull request to explicitly add BLIS support under the sensible
        # subclass name "blis_info" (see Numpy PR #7294), explicitly listing
        # BLIS here should assist in future-proofing our multithreading
        # detection.
        #r'blis',
    ))

    # Regular expression fragment matching uniquely identifying substrings of
    # dirnames of optimized BLAS shared libraries.
    blas_lib_dirname_regex = r'|'.join((
        # AMD Core Math Library (ACML).
        r'acml',

        # Automatically Tuned Linear Algebra Software (ATLAS).
        r'atlas',

        # Intel Math Kernel Library (MKL).
        r'mkl',

        # OpenBLAS.
        r'openblas',
    ))

    # Regular expression fragment matching the boundary of a dirname at which a
    # substring matching an optimized BLAS substring may either begin or end.
    dirname_boundary_regex = r'[{}_.-]'.format(dirs.SEPARATOR_REGEX)

    #FIXME: What about macOS? Are shared libraries prefixed by "lib" under that
    #platform as well?

    # Redefine this global.
    _OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX = (
        r'^lib({})(?:[_.-].*)?$'.format(blas_lib_basename_regex))

    # Redefine this global. Since Numpy has already stripped all
    # platform-specific prefixes (e.g., "lib") and suffixes (e.g., ".so") from
    # this basename, only this substring followed by an arbitrary bounded suffix
    # need be matched.
    _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX = (
        r'^({})(?:[_-].*)?$'.format(blas_lib_basename_regex))

    # Redefine these globals to the same regular expression.
    _OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX = (
        r'^.*?{boundary}({blas_dirname})(?:{boundary}.*)?$'.format(
            boundary=dirname_boundary_regex,
            blas_dirname=blas_lib_dirname_regex,))
    _OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX = (
        _OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX)

# ....................{ TESTERS                            }....................
def is_blas_optimized() -> bool:
    '''
    ``True`` only" if the currently installed version of Numpy is linked against an
    optimized BLAS (Basic Linear Algebra Subprograms) implementation, ideally
    but _not_ necessarily parallelized across multiple processors.

    Optimized BLAS implementations are _strongly_ recommended over unoptimized
    BLAS implementations. The `numpy.dot()` operator, which is implicitly
    optimized when Numpy is linked against a optimized BLAS implementation, is
    frequently called by BETSE in its critical path.

    Note that testing for parallelized optimized BLAS implementations, while
    more specific and hence preferable, is infeasible for common edge-cases
    (e.g., Debian-based Linux distributions). For further details, see the
    :data:`_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX` string global.
    '''

    # For each private tester implementing a heuristic for this public test (in
    # order of decreasing generality, portability, and reliability)...
    for tester_heuristic in (
        _is_blas_optimized_opt_info_libraries,
        _is_blas_optimized_opt_info_library_dirs,
        _is_blas_optimized_opt_info_macos,
        _is_blas_optimized_posix_symlink,
    ):
        # Attempt to...
        try:
            # Log the current heuristic being attempted.
            logs.log_debug(
                'Detecting BLAS by heuristic %s()...',
                tester_heuristic.__name__)

            # Call this tester, capturing the result for subsequent handling.
            tester_result = tester_heuristic()

            # If this tester definitively identified Numpy as either
            # optimized or non-optimized...
            if tester_result is not None:
                # Log this result.
                logs.log_debug('BLAS optimization detected: %r', tester_result)

                # Return this result.
                return tester_result
            # Else, continue to the next tester.
        # If an error occurs, log that error *WITHOUT* raising an exception.
        # Detecting Numpy optimization is non-essential and hence hardly worth
        # halting the application over.
        except Exception as exception:
            logs.log_exception(exception)

    # Else, all heuristics failed to definitively identify Numpy to be either
    # optimized or non-optimized. For safety, assume the latter.
    return False

# ....................{ TESTERS ~ private : opt_info       }....................
def _is_blas_optimized_opt_info_libraries() -> BoolOrNoneTypes:
    '''
    ``True`` only" if the first element of the `libraries` list of the global
    :data:`numpy.__config__.blas_opt_info` dictionary heuristically
    corresponds to that of an optimized BLAS implementation, `False` if a non-
    fatal error condition arises (e.g., due this list or dictionary being
    undefined), _or_ `None` otherwise.

    This function returns `None` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.

    Numpy does _not_ define a public API directly defining this boolean. Numpy
    does, however, define a private API defining a variety of metadata from
    which this boolean is indirectly derivable: the :mod:`numpy.__config__`
    submodule. The :func:`numpy.distutils.misc_util.generate_config_py` function
    programmatically fabricates the contents of the :mod:`numpy.__config__`
    submodule at Numpy installation time. This function introspectively inspects
    these contents for uniquely identifying metadata in a portable manner.
    '''

    # Global BLAS linkage dictionary for this Numpy installation if any or
    # "None" otherwise. Technically, this dictionary should *ALWAYS* be defined.
    # Reality probably occasionally begs to disagree, however.
    blas_opt_info = getattr(numpy_config, 'blas_opt_info', None)

    # If this dictionary is undefined, log a non-fatal warning and return False.
    # While unfortunate, this is *NOT* worth raising a fatal exception over.
    if blas_opt_info is None:
        logs.log_warning(
            'Numpy installation misconfigured: '
            '"numpy.__config__.blas_opt_info" dictionary not found.')
        return False

    # List of the uniquely identifying substrings of all BLAS library basenames
    # this version of Numpy is linked against in a high-level manner if any or
    # "None" otherwise.
    #
    # Note that this list is *NOT* guaranteed to exist. When this version of
    # Numpy is linked against a BLAS library in a low-level manner (e.g., via
    # "'extra_link_args': ['-Wl,-framework', '-Wl,Accelerate']" on macOS), this
    # list should *NOT* exist. In most other cases, this list should exist. To
    # avoid edge cases, this list is ignored if absent.
    blas_basename_substrs = blas_opt_info.get('libraries', None)

    # If this list is either undefined or empty, silently noop.
    if not blas_basename_substrs:
        return None
    # Else, this list is non-empty.

    # First element of this list. For simplicity, this function assumes the
    # BLAS library identified by this element currently exists. While
    # iteratively testing all listed BLAS libraries for existence would be
    # feasible, doing so is platform-specific and hence non-trivially fragile.
    blas_basename_substr = blas_basename_substrs[0]

    # If the BLAS library identified by this element is optimized, return
    # True. Since this element may be suffixed by non-identifying metadata
    # (e.g., version), a regular expression is leveraged.
    if regexes.is_match(
        text=blas_basename_substr,
        regex=_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_REGEX,
    ):
        return True

    # Else, instruct our caller to continue to the next heuristic.
    return None


def _is_blas_optimized_opt_info_library_dirs() -> BoolOrNoneTypes:
    '''
    ``True`` only" if the first element of the `library_dirs` list of the global
    :data:`numpy.__config__.blas_opt_info` dictionary heuristically
    corresponds to that of an optimized BLAS implementation, `False` if a non-
    fatal error condition arises (e.g., due this list or dictionary being
    undefined), _or_ `None` otherwise.

    This function returns `None` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.
    '''

    # List of the dirnames of all BLAS libraries this version of Numpy is linked
    # against in a high-level manner if any or "None" otherwise.
    #
    # Note that the "blas_opt_info" dictionary global is guaranteed to exist due
    # to the previously called _is_blas_optimized_opt_info_basename() function.
    #
    # Note that this list is *NOT* guaranteed to exist. When this version of
    # Numpy is linked against a BLAS library in a low-level manner (e.g., via
    # "'extra_link_args': ['-Wl,-framework', '-Wl,Accelerate']" on macOS), this
    # list should *NOT* exist. In most other cases, this list should exist. To
    # avoid edge cases, this list is ignored if absent.
    blas_dirnames = numpy_config.blas_opt_info.get('library_dirs', None)

    # If this list is either undefined or empty, silently noop.
    if not blas_dirnames:
        return None
    # Else, this list is non-empty.

    # First element of this list. For simplicity, this function assumes the
    # BLAS library identified by this element currently exists. While
    # iteratively testing all listed BLAS libraries for existence would be
    # feasible, doing so is platform-specific and hence non-trivially fragile.
    blas_dirname = blas_dirnames[0]

    # If the BLAS library identified by this element is optimized, return
    # True. Since this element is an arbitrary pathname typically containing
    # non-identifying metadata, a regular expression is leveraged.
    if regexes.is_match(
        text=blas_dirname,
        regex=_OPTIMIZED_BLAS_OPT_INFO_LIBRARY_DIRS_REGEX,
    ):
        return True

    # Else, instruct our caller to continue to the next heuristic.
    return None


def _is_blas_optimized_opt_info_macos() -> BoolOrNoneTypes:
    '''
    ``True`` only" if the current platform is macOS *and* the
    ``extra_link_args`` list of the global
    :data:`numpy.__config__.blas_opt_info` dictionary both exists *and*
    heuristically corresponds to that of an optimized BLAS implementation
    specific to macOS (e.g., Accelerate, vecLib), ``False`` if a non-fatal error
    condition arises (e.g., due this list or dictionary being undefined), *or*
    ``None`` otherwise.

    This function returns ``None`` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.

    Unlike all other BLAS implementations, macOS-specific BLAS implementations
    are linked against with explicit linker flags rather than pathnames. For
    further confirmation that the :attr:`numpy.__config__.blas_opt_info`
    dictionary defines these flags when linked to these implementations, see:

    * https://trac.macports.org/ticket/22200
    * https://github.com/BVLC/caffe/issues/2677

    When life buys you cat food, you eat cat food.
    '''

    # If the current platform is *NOT* macOS, continue to the next heuristic.
    if not oses.is_macos():
        return None
    # Else, the current platform is macOS.

    # List of all implementation-specific link arguments with which Numpy linked
    # against the current BLAS implementation if any or "None".
    #
    # Note that the "blas_opt_info" dictionary global is guaranteed to exist due
    # to the previously called _is_blas_optimized_opt_info_basename() function.
    blas_link_args_list = numpy_config.blas_opt_info.get(
        'extra_link_args', None)

    # If no such argument exists, continue to the next heuristic. Since this
    # list is strictly optional, no errors or warnings are logged.
    if not blas_link_args_list:
        return None

    # Set of these arguments, converted from this list for efficiency.
    blas_link_args = set(blas_link_args_list)
    # logs.log_info('blas_link_args: {}'.format(blas_link_args))

    # Subset of this set specific to multithreaded BLAS implementations.
    blas_link_args_multithreaded = (
        blas_link_args & _OPTIMIZED_BLAS_OPT_INFO_EXTRA_LINK_ARGS_MACOS)

    # If this subset is nonempty, return True.
    if len(blas_link_args_multithreaded) > 0:
        return True

    # Else, instruct our caller to continue to the next heuristic.
    return None

# ....................{ TESTERS ~ private : linkage        }....................
def _is_blas_optimized_posix_symlink() -> BoolOrNoneTypes:
    '''
    ``True`` only" if the current platform is POSIX-compliant and hence supports
    symbolic links _and_ the first element of the `libraries` list of the global
    :data:`numpy.__config__.blas_opt_info` dictionary is a symbolic link
    masquerading as either the unoptimized reference BLAS implementation but in
    fact linking to an optimized BLAS implementation.

    This function returns `None` when unable to deterministically decide this
    boolean, in which case a subsequent heuristic will attempt to do so.
    '''

    # If the current platform is *NOT* POSIX-compliant and hence does *NOT*
    # support symbolic links, continue to the next heuristic.
    if not oses.is_posix():
        return None

    #FIXME: Generalize to macOS as well once the
    #libs.iter_linked_filenames() function supports macOS.

    # If the current platform is *NOT* Linux, continue to the next heuristic.
    #
    # The libs.iter_linked_filenames() function called below currently only
    # supports Linux.
    if not oses.is_linux():
        return None

    # First element of the list of uniquely identifying substrings of all BLAS
    # library basenames this version of Numpy is linked against.
    #
    # Note that this list is guaranteed to both exist and be non-empty due to
    # the previously called _is_blas_optimized_opt_info_basename() function.
    blas_basename_substr = numpy_config.blas_opt_info['libraries'][0]

    # If this element appears to be neither the reference BLAS or CBLAS
    # implementations (e.g., "blas", "cblas", "refblas", "refcblas"), continue
    # to the next heuristic.
    if not blas_basename_substr.endswith('blas'):
        return None

    # Absolute path of an arbitrary shared library-based Numpy C extension.
    #
    # Unfortunately, the "numpy.__config__" API fails to specify the absolute
    # paths of the libraries it links against. Since there exists no reliable
    # means of reverse engineering these paths from this API, these paths must
    # be obtained by another means: specifically, by querying the standard
    # "numpy.core.multiarray" C extension installed under all supported Numpy
    # for the absolute paths of all external shared libraries to which this
    # extension links -- exactly one of which is guaranteed to be the absolute
    # path of what appears to be a reference BLAS or CBLAS implementation.
    numpy_lib_filename = modules.get_filename('numpy.core.multiarray')

    # For the basename and absolute path of each shared library linked to
    # by this Numpy shared library...
    for (numpy_linked_lib_basename, numpy_linked_lib_filename) in (
        dlls.iter_linked_filenames(numpy_lib_filename)):
        # Basename excluding all suffixing filetypes of this library.
        numpy_linked_lib_rootname = pathnames.get_pathname_sans_filetypes(
            numpy_linked_lib_basename)
        # print('rootname: {}; basename: {}; filename: {}'.format(numpy_linked_lib_rootname, numpy_linked_lib_basename, numpy_linked_lib_filename))

        # If this appears to be neither the BLAS nor CBLAS reference library,
        # continue to the next library.
        if not numpy_linked_lib_rootname.endswith('blas'):
            continue
        # Else, this is either the BLAS or CBLAS reference library.

        # If this library is *NOT* a symbolic link, Numpy links directly against
        # an unoptimized BLAS implementation. Halt!
        if not files.is_symlink(numpy_linked_lib_filename):
            break
        # Else, this library is actually a symbolic link to another library.

        # Absolute path of the target library to which this library links.
        numpy_linked_lib_target_filename = pathnames.canonicalize(
            numpy_linked_lib_filename)
        # print('target filename: {}'.format(numpy_linked_lib_target_filename))

        # If either the basename or dirname of this path corresponds to that of
        # an optimized BLAS library, return True.
        if regexes.is_match(
            text=pathnames.get_basename(numpy_linked_lib_target_filename),
            regex=_OPTIMIZED_BLAS_LINKED_LIB_BASENAME_REGEX,
        ) or regexes.is_match(
            text=pathnames.get_dirname(numpy_linked_lib_target_filename),
            regex=_OPTIMIZED_BLAS_LINKED_LIB_DIRNAME_REGEX,
        ):
            return True

        # Else, Numpy links against an unoptimized BLAS implementation. Halt!
        break

    # Else, instruct our caller to continue to the next heuristic.
    return None

# ....................{ GETTERS                            }....................
def get_metadatas() -> tuple:
    '''
    Tuple of 2-tuples ``(metedata_name, metadata_value)``, describing all
    currently installed third-party dependencies against which Numpy was linked
    (e.g., BLAS, LAPACK).
    '''

    #FIXME: Add LAPACK linkage metadata as well.
    return (
        ('numpy (blas)', get_blas_metadata()),
    )


def get_blas_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the current Numpy installation with respect
    to BLAS linkage.
    '''

    # This dictionary.
    metadata = OrderedArgsDict('optimized', is_blas_optimized())

    # Set of all keys of the dictionary global synopsizing this metadata,
    # sorted in ascending lexicographic order for readability.
    blas_opt_info_keys = iterables.sort_ascending(
        tuple(numpy_config.blas_opt_info.keys()))

    # For each such key...
    for blas_opt_info_key in blas_opt_info_keys:
        # The value of this key, unconditionally converted into a string and
        # then trimmed to a reasonable string length. The values of numerous
        # keys (e.g., "libraries", "sources") commonly exceed this length,
        # hampering readability for little to no gain. Excise them all.
        metadata[blas_opt_info_key] = strs.trim(
            obj=numpy_config.blas_opt_info[blas_opt_info_key],
            max_len=256,
        )

    # Return this dictionary.
    return metadata



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Fixtures and fixture classes creating temporary simulation configurations
isolated to specific tests, which typically modify the contents of these
configurations so as to exercise specific feature sets and edge cases.
'''

# ....................{ IMPORTS                            }....................
from betse_test.fixture.simconf.simconfclser import (
    SimConfTestExternal, SimConfTestInternal)
from pytest import fixture
from py._path.local import LocalPath

# ....................{ FIXTURES                           }....................
# Test-scope fixture creating and returning a new object for each discrete test.
@fixture
def betse_sim_conf(betse_temp_dir: LocalPath) -> SimConfTestInternal:
    '''
    Per-test fixture creating a temporary minified simulation configuration file
    and returning an object encapsulating the contents of this file.

    Configuration Modifications (On-disk)
    ----------
    This fixture copies BETSE's default simulation configuration file,
    complete with all external assets (e.g., geometry masks) referenced and
    required by this file, into a temporary directory whose basename is the name
    of the test requesting this fixture excluding the prefixing substring
    ``test_``. When requested by the ``test_cli_sim_default`` test, for example,
    this fixture creates a temporary simulation configuration file
    ``{tmpdir}/cli_sim_default/sim_config.yaml`` for the absolute path
    ``{tmpdir}`` of this test session's root temporary directory (e.g.,
    ``/tmp/pytest-0/cli_sim_default/sim_config.yaml``).

    This directory and hence simulation configuration is safely accessible
    *only* for the duration of the current test. Subsequently run tests and
    fixtures *cannot* safely reuse this configuration.

    Configuration Modifications (In-memory)
    ----------
    This fixture also transforms the in-memory instance of the
    :class:`betse.science.parameters.Parameters` class encapsulating this
    configuration as follows:

    * All configuration options either requiring interactive input *or*
      displaying interactive output are disabled (e.g., plots, animations).
    * The space and time costs associated with simulating this configuration
      are safely minimized in a manner preserving all features.

    Since this fixture does *not* write these changes back to this file, the
    parent fixture or test is expected to do so manually (e.g., by calling the
    :meth:`SimConfTestInternal.config.overwrite` method on the object returned
    by this fixture).

    Parameters
    ----------
    betse_temp_dir : LocalPath
        Object encapsulating a temporary directory isolated to the current test.

    Returns
    ----------
    SimConfTestInternal
        Test-specific object encapsulating a temporary simulation configuration
        file specific to the current test, including such metadata as:
        * The absolute path of this configuration's on-disk YAML file.
        * This configuration's in-memory dictionary deserialized from this file.
    '''

    # Absolute path of this configuration file in this temporary directory.
    sim_conf_filepath = betse_temp_dir.join('sim_config.yaml')

    # Test-specific object encapsulating this simulation configuration file.
    sim_state = SimConfTestInternal(conf_filepath=sim_conf_filepath)

    # Minimize the space and time costs associated with this configuration.
    sim_state.config.minify()

    # Return this object.
    return sim_state


# Test-scope fixture creating and returning a new object for each discrete test.
@fixture
def betse_sim_conf_default(betse_temp_dir: LocalPath) -> SimConfTestInternal:
    '''
    Per-test fixture creating a temporary default simulation configuration file
    and returning an object encapsulating the contents of this file.

    Unlike the minified simulation configuration created by the
    :func:`betse_sim_conf` fixture and leveraged by most tests, the default
    simulation configuration created by this fixture and leveraged by few tests
    is unmodified (except for disabling interactive simulation features, which
    non-interactive testing unavoidably requires). Tests doing so incur a
    significant performance penalty but can expose edge-case issues obscured by
    minification, including computational instability produced by the default
    non-minified time steps.

    Parameters
    ----------
    betse_temp_dir : LocalPath
        Object encapsulating a temporary directory isolated to the current test.

    Returns
    ----------
    SimConfTestInternal
        Test-specific object encapsulating a temporary simulation configuration.

    See Also
    ----------
    :func:`betse_sim_conf`
        Further details, ignoring minification performed by this fixture.
    '''

    # Absolute path of this configuration file in this temporary directory.
    sim_conf_filepath = betse_temp_dir.join('sim_config.yaml')

    # Test-specific object encapsulating this simulation configuration file.
    sim_state = SimConfTestInternal(conf_filepath=sim_conf_filepath)

    # Return this object *WITHOUT* calling sim_state.config.minify().
    return sim_state


@fixture
def betse_sim_conf_compat(
    betse_temp_dir: LocalPath) -> SimConfTestExternal:
    '''
    Per-test fixture creating and returning an object encapsulating a temporary
    simulation configuration file (complete with a pickled seed, initialization,
    and simulation) produced by the oldest version of this application for which
    the current version of this application guarantees backward compatibility.

    Caveats
    ----------
    Unlike the object returned by the comparable :func:`betse_sim_conf` fixture,
    the object returned by this fixture is *not* safely modifiable by the
    current version of this application. Doing so would invalidate the pickled
    files produced by the older version of this application, which would largely
    defeat the purpose of invoking this fixture.

    Parameters
    ----------
    betse_temp_dir : LocalPath
        Object encapsulating a temporary directory isolated to the current test.

    Returns
    ----------
    SimConfTestExternal
        Test-specific object encapsulating a temporary simulation configuration
        file specific to the current test, complete with pickled seed,
        initialization, and simulation files produced by the older version of
        this application.
    '''

    # Defer heavyweight imports.
    from betse import metadata, pathtree
    from betse.util.io.log import logs
    from betse.util.os.shell import shelldir
    from betse.util.path import gits
    from betse.util.path.command import cmdrun
    from betse.util.py import pys

    # Character uniquely padding all banners logged below.
    BANNER_PADDING = '*'

    # Absolute path of the top-level temporary directory isolated to this test.
    betse_temp_dirname = str(betse_temp_dir)

    # ..................{ PHASE                              }..................
    # Log a single-line terminal banner identifying the initial fixture phase.
    logs.log_banner(title='PHASE 1: shallow git clone', padding=BANNER_PADDING)

    # Absolute pathname of this application's Git-based working tree. Since this
    # test suite should only every be run from within a working tree, this
    # retrieval should *ALWAYS* succeed.
    git_worktree_dirname = pathtree.get_git_worktree_dirname_or_none()

    # Absolute path of a temporary non-existing directory isolated to this test
    # to clone the older version of this application into.
    betse_old_dirpath = betse_temp_dir.join('betse_old')
    betse_old_dirname = str(betse_old_dirpath)

    # Shallowly clone from the tag referring to the older version of this
    # application in this Git working tree into this temporary directory.
    gits.clone_worktree_shallow(
        branch_or_tag_name=metadata.GIT_TAG_OLDEST_BACKWARD_COMPATIBILITY,
        src_dirname=git_worktree_dirname,
        trg_dirname=betse_old_dirname,
    )

    # ..................{ PHASE                              }..................
    # Log a single-line terminal banner identifying the next fixture phase.
    logs.log_banner(title='PHASE 2: sim config export', padding=BANNER_PADDING)

    # Name of the functional test exporting an obsolete simulation configuration
    # from this older version.
    test_cli_sim_export_name = 'test_cli_sim_export'

    # Absolute path of the simulation configuration file exported by this test
    # into a temporary directory isolated to the current test.
    sim_conf_old_filepath = betse_temp_dir.join(
        test_cli_sim_export_name, 'sim_config.yaml')
    sim_conf_old_filename = str(sim_conf_old_filepath)

    # List of one or more shell words unambiguously running the executable
    # specific to the active Python interpreter and machine architecture.
    py_command_line_prefix = pys.get_command_line_prefix()

    # List of shell words comprising the "py.test"-based command exporting this
    # old simulation configuration.
    export_sim_conf_old_command = py_command_line_prefix + [
        'setup.py', 'test',
        '-k', test_cli_sim_export_name,
        '--export-sim-conf-dir', betse_temp_dirname,
    ]

    # Temporary change to the directory containing this "setup.py" script.
    with shelldir.setting_cwd(betse_old_dirname):
        # Export this old simulation configuration with this script.
        cmdrun.run_or_die(command_words=export_sim_conf_old_command)

    # Test-specific object encapsulating this simulation configuration file.
    sim_state = SimConfTestExternal(conf_filename=sim_conf_old_filename)

    # ..................{ PHASE                              }..................
    # Log a single-line terminal banner identifying the final fixture phase.
    logs.log_banner(title='PHASE 3: sim config test', padding=BANNER_PADDING)

    # Return this object.
    return sim_state



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level string facilities.
'''

# ....................{ IMPORTS                            }....................
import html, textwrap
from betse.exceptions import BetseStrException
from betse.util.type import types
from betse.util.type.types import type_check, IterableTypes, StrOrNoneTypes
from textwrap import TextWrapper

# For convenience, permit callers to import the general-purpose trim() function
# from this submodule rather than the "types" submodule.
from betse.util.type.types import trim

if False: trim  # silence IDE warnings

# ....................{ SINGLETONS                         }....................
text_wrapper = TextWrapper()
'''
Singleton :class:`TextWrapper` instance with which to wrap text.

This singleton improves efficiency -- occasionally dramatically. *All* public
functions provided by the :mod:`textwrap` module implicitly instantiate
temporary :class:`TextWrapper` instances on each function call.
'''

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_if_empty(text: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a human-readable
    message) if the passed string is empty.

    Parameters
    ----------
    text : str
        String to be validated.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseStrException
        If this string is empty.
    '''

    # If this string is empty, raise an exception.
    if not text:
        # If no exception message was passed, synthesize one.
        if not exception_message:
            exception_message = 'Text empty.'.format()

        # Raise this exception.
        raise BetseStrException(exception_message)


@type_check
def die_unless_prefix(
    text: str, prefix: str, exception_message: StrOrNoneTypes = None) -> None:
    '''
    Raise an exception with the passed message (defaulting to a message
    synthesized from the passed arguments) if the second passed string does
    *not* prefix the first passed string.

    Parameters
    ----------
    text : str
        String to be validated.
    prefix : str
        Prefix to test for.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Raises
    ----------
    BetseStrException
        If this string is *not* prefixed by this prefix.
    '''

    # If this string is *NOT* prefixed by this prefix, raise an exception.
    if not is_prefix(text, prefix):
        # If no exception message was passed, synthesize one from this name.
        if not exception_message:
            exception_message = 'Text "{}" not prefixed by "{}".'.format(
                text, prefix)

        # Raise this exception.
        raise BetseStrException(exception_message)

# ....................{ TESTERS ~ case                     }....................
@type_check
def is_lowercase(text: str) -> bool:
    '''
    ``True`` only if the passed string is strictly lowercase (i.e., this string
    contains at least one alphabetic character *and* all such characters are
    lowercase).

    This function ignores all non-alphabetic characters (e.g., digits,
    punctuation) in this string.
    '''

    # O.K., that's pretty sweet.
    return text.islower()


@type_check
def is_uppercase(text: str) -> bool:
    '''
    ``True`` only if the passed string is strictly uppercase (i.e., this string
    contains at least one alphabetic character *and* all such characters are
    uppercase).

    This function ignores all non-alphabetic characters (e.g., digits,
    punctuation) in this string.
    '''

    # Ditto on the sweetness.
    return text.isupper()

# ....................{ TESTERS ~ [pre|suf]fix             }....................
def is_prefix(text: str, prefix: str) -> bool:
    '''
    ``True`` only if the second passed string prefixes the first passed string.

    Parameters
    ----------
    text : str
        String to be tested.
    prefix : str
        Prefix to test for.
    '''

    return text.startswith(prefix)


@type_check
def is_suffix(text: str, suffix: str) -> bool:
    '''
    ``True`` only if the second passed string suffixes the first passed string.

    Parameters
    ----------
    text : str
        String to be tested.
    suffix : str
        Prefix to test for.
    '''

    return text.endswith(suffix)

# ....................{ ADDERS                             }....................
def add_prefix_unless_found(text: str, prefix: str) -> str:
    '''
    Prefix the passed string by the passed prefix unless the former is already
    prefixed by the latter.

    Parameters
    ----------
    text : str
        String to add this prefix to.
    prefix : str
        Prefix to add to this string.

    Returns
    ----------
    str
        The former prefixed by the latter.
    '''

    return text if is_prefix(text, prefix) else prefix + text


def add_suffix_unless_found(text: str, suffix: str) -> str:
    '''
    Suffix the passed string by the passed suffix unless the former is already
    suffixed by the latter.

    Parameters
    ----------
    text : str
        String to add this suffix to.
    suffix : str
        Suffix to add to this string.

    Returns
    ----------
    str
        The former suffixed by the latter.
    '''

    return text if is_suffix(text, suffix) else text + suffix

# ....................{ JOINERS ~ on                       }....................
def join(*texts) -> str:
    '''
    Concatenation of the passed strings with no separating delimiter.

    This is a convenience function wrapping the standard ``"".join((...))``
    method, whose syntax is arguably overly verbose.
    '''

    return join_on(*texts, delimiter='')


@type_check
def join_by_index(iterable: IterableTypes, subiterable_index: object) -> str:
    '''
    Concatenation of each string at the passed key or index of each subiterable
    of the passed iterable, with no separating delimiter.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable of subiterables to be summed.
    subiterable_index : object
        Object with which to index each subiterable of this iterable. The type
        of this object *must* be a type accepted by the ``__getitem__()``
        special method of each subiterable. If each subiterable is a:
        * **Mapping** (e.g., :class:`dict`), this object *must* be hashable.
        * **Sequence** (e.g., :class:`list`, :class:`tuple`), this object
          *must* be either:
          * An integer.
          * A :func:`slice` object.

    Returns
    ----------
    str
        Concatenation of each string at this key or index of each subiterable of
        this iterable, with no separating delimiter.
    '''

    # Efficiency and simplicity combine here to form SuperHappyFunFunction.
    return ''.join(subiterable[subiterable_index] for subiterable in iterable)

# ....................{ JOINERS ~ on                       }....................
def join_on_newline(*texts) -> str:
    '''
    Join the passed strings with newline as the separating delimiter.

    This is a convnience function wrapping the standard
    ``"\n".join((...))`` method, whose syntax is arguably overly verbose.
    '''

    return join_on(*texts, delimiter='\n')


def join_on_space(*texts) -> str:
    '''
    Join the passed strings with space as the separating delimiter.

    This is a convnience function wrapping the standard
    ``" ".join((...))`` method, whose syntax is arguably overly verbose.
    '''

    return join_on(*texts, delimiter=' ')


@type_check
def join_on(*texts: IterableTypes, delimiter: str) -> str:
    '''
    Join the passed strings with the passed separating delimiter.

    This is a convenience function wrapping the standard
    ``"...".join((...))`` method, whose syntax is arguably overly obfuscated.

    Parameters
    ----------
    texts : SequenceTypes
        Tuple of all strings to be joined, consisting of either:
        * One or more strings.
        * One iterable of strings.
    delimiter : str
        Substring to join each such string on.

    Returns
    ----------
    str
        String joined from the passed strings with the passed delimiter.
    '''

    # If only one object was passed and this object is a non-string iterable
    # (e.g, list, tuple), default the list of passed strings to this object.
    if len(texts) == 1 and types.is_iterable_nonstr(texts[0]):
        texts = texts[0]

    # Join these strings.
    return delimiter.join(texts)

# ....................{ JOINERS ~ as                       }....................
@type_check
def join_as(
    *texts,
    delimiter_if_two: str,
    delimiter_if_three_or_more_nonlast: str,
    delimiter_if_three_or_more_last: str
) -> str:
    '''
    Join the passed strings in a human-readable manner.

    If:

    * No strings are passed, the empty string is returned.
    * One string is passed, this string is returned as is without modification.
    * Two strings are passed, these strings are joined with the passed
      `delimiter_if_two` separator.
    * Three or more strings are passed:
      * All such strings except the last two are joined with the passed
        `delimiter_if_three_or_more_nonlast` separator.
      * The last two such strings are joined with the passed
        `delimiter_if_three_or_more_last` separator.

    Parameters
    ----------
    texts : Tuple[str]
        List of all strings to be joined.
    delimiter_if_two : str
        String separating each element of `texts` if `len(texts) == 2`.
    delimiter_if_three_or_more_nonlast : str
        String separating each element _except_ the last two of `texts` if
        `len(texts) >= 3`.
    delimiter_if_three_or_more_last : str
        String separating the last two elements of `texts` if `len(texts) >= 3`.

    Returns
    ----------
    str
        Resulting string as described above.

    Examples
    ----------
    >>> join_as(
    ...     ('Fulgrim', 'Perturabo', 'Angron', 'Mortarion'),
    ...     delimiter_if_two=' and ',
    ...     delimiter_if_three_or_more_nonlast=', ',
    ...     delimiter_if_three_or_more_last=', and '
    ... )
    'Fulgrim, Perturabo, Angron, and Mortarion'
    '''

    # Number of passed strings.
    texts_count = len(texts)

    # If no strings are passed, return the empty string.
    if texts_count == 0:
        return ''
    # If one string is passed, return this string as is.
    elif texts_count == 1:
        return texts[0]
    # If two strings are passed, return these strings joined appropriately.
    elif texts_count == 2:
        return '{}{}{}'.format(
            texts[0], delimiter_if_two, texts[1])
    # If three or more strings are passed, return these strings joined
    # appropriately.
    else:
        # All such strings except the last two, joined appropriately.
        texts_nonlast = join_on(
            *texts[0:-2], delimiter=delimiter_if_three_or_more_nonlast)

        # The last two such strings, joined appropriately.
        texts_last = '{}{}{}'.format(
            texts[-2], delimiter_if_three_or_more_last, texts[-1])

        # Return these two substrings, joined appropriately.
        return '{}{}{}'.format(
            texts_nonlast, delimiter_if_three_or_more_nonlast, texts_last)

# ....................{ JOINERS ~ as : conjunction         }....................
@type_check
def join_as_conjunction(*texts: str) -> str:
    '''
    Conjunctively join all passed strings in a human-readable manner.

    Specifically:

    * All passed strings excluding the last two are joined with `, `.
    * The last two passed strings are joined with `, and `.
    '''

    return join_as(
        *texts,
        delimiter_if_two=' and ',
        delimiter_if_three_or_more_nonlast=', ',
        delimiter_if_three_or_more_last=', and '
    )


@type_check
def join_as_conjunction_double_quoted(*texts: str) -> str:
    '''
    Conjunctively double-quote and join all passed strings in a human-readable
    manner.

    Specifically:

    * All passed strings are double-quoted.
    * All passed strings excluding the last two are joined with `, `.
    * The last two passed strings are joined with `, and `.
    '''

    # Tuple of all passed strings double-quoted. Since the "*" operator applied
    # to this tuple below requires a sequence rather than generator, this is the
    # most space-efficient available sequence (i.e., frozen tuple).
    texts_quoted = tuple(double_quote(text) for text in texts)

    # Conjunctively join these strings.
    return join_as_conjunction(*texts_quoted)

# ....................{ JOINERS ~ as : conjunction         }....................
@type_check
def join_as_disjunction(*texts: str) -> str:
    '''
    Disjunctively join all passed strings in a human-readable manner.

    Specifically:

    * All passed strings excluding the last two are joined with `, `.
    * The last two passed strings are joined with `, or `.
    '''

    return join_as(
        *texts,
        delimiter_if_two=' or ',
        delimiter_if_three_or_more_nonlast=', ',
        delimiter_if_three_or_more_last=', or '
    )


@type_check
def join_as_disconjunction_double_quoted(*texts: str) -> str:
    '''
    Disjunctively double-quote and join all passed strings in a human-readable
    manner.

    Specifically:

    * All passed strings are double-quoted.
    * All passed strings excluding the last two are joined with `, `.
    * The last two passed strings are joined with `, or `.
    '''

    # Tuple of all passed strings double-quoted. See
    # join_as_conjunction_double_quoted().
    texts_quoted = tuple(double_quote(text) for text in texts)

    # Disjunctively join these strings.
    return join_as_disjunction(*texts_quoted)

# ....................{ QUOTERS                            }....................
#FIXME: We don't actually escape embedded double quotes yet. The reason why is
#that doing is subtly non-trivial; for safety, we don't want to re-escape
#already escaped double quotes in this string. We only want to escape
#currently unescaped double quotes in this string. To do so, we'll need to
#leverage regex-based substitution with negative lookbehind preventing existing
#substrings matching '\\"' from being subject to substitution.

def double_quote(text: str) -> str:
    '''
    Double-quote the passed string in a human-readable manner.

    Specifically (in order):

    . All prefixing and suffixing whitespace is removed from this string.
    . If this string is either double- or single-quoted, these quotes are
      removed.
    . All other double quotes in this string are escaped (e.g., replacing each
      `"` character with `\\"`).
    . The resulting string is double-quoted and returned.
    '''

    # Remove all prefixing and suffixing whitespace from this string.
    text = remove_whitespace_presuffix(text)

    # If this string is already either double- or single-quoted, strip these
    # delimiting quotes for simplicity.
    if ((text[0] == '"' and text[-1] == '"') or
        (text[0] == "'" and text[-1] == "'")):
        text = text[1:-1]

    # Double quote this string.
    return '"{}"'.format(text)

# ....................{ REMOVERS ~ newline                 }....................
@type_check
def remove_newlines_suffix(text: str) -> str:
    '''
    Passed string with all suffixing (but *not* prefixing) newlines removed.
    '''

    return text.rstrip('\n')

# ....................{ REMOVERS ~ space                   }....................
@type_check
def remove_whitespace(text: str) -> str:
    '''
    Passed string with *all* whitespace removed, including all prefixing and
    suffixing whitespace and whitespace interspersed through this string.

    Parameters
    ----------
    text : str
        String to remove all whitespace from.

    Returns
    ----------
    str
        String with all whitespace removed.

    See Also
    ----------
    https://stackoverflow.com/a/8270124/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # While a regular expression-based implementation is also feasible, the
    # current approach of splitting this string on whitespace into substrings
    # containing no whitespace and then concatenating these substrings is
    # substantially more Pythonic (and presumably efficient).
    return ''.join(text.split())


@type_check
def remove_whitespace_presuffix(text: str) -> str:
    '''
    Passed string with all prefixing and suffixing whitespace removed.
    '''

    return text.strip()


@type_check
def remove_whitespace_suffix(text: str) -> str:
    '''
    Passed string with all suffixing (but *not* prefixing) whitespace removed.
    '''

    return text.rstrip()

# ....................{ REMOVERS ~ prefix                  }....................
@type_check
def remove_prefix(text: str, prefix: str, exception_message: str = None) -> str:
    '''
    Passed string with the passed prefix removed if present *or* raise an
    exception with the passed message otherwise.

    Parameters
    ----------
    text : str
        String to be examined. Since strings are immutable in Python, this
        string remains unmodified.
    prefix : str
        Prefix to remove from this string.
    exception_message : optional[str]
        Exception message to be raised. Defaults to ``None``, in which case an
        exception message synthesized from the passed arguments is raised.

    Returns
    ----------
    str
        Resulting string as described above.
    '''

    # If this string is *NOT* prefixed by this prefix, raise this exception.
    die_unless_prefix(text, prefix, exception_message)

    # Else, return a new string with this prefix removed from this string.
    return remove_prefix_if_found(text, prefix)


@type_check
def remove_prefix_if_found(text: str, prefix: str) -> str:
    '''
    Passed string with the passed prefix removed if present *or* the passed
    string as is otherwise.

    Parameters
    ----------
    text : str
        String to be examined. Since strings are immutable in Python, this
        string remains unmodified.
    prefix : str
        Prefix to remove from this string.

    Returns
    ----------
    str
        Resulting string as described above.
    '''

    return text[len(prefix):] if is_prefix(text, prefix) else text

# ....................{ REMOVERS ~ suffix                  }....................
@type_check
def remove_suffix_if_found(text: str, suffix: str) -> str:
    '''
    Passed string with the passed suffix removed if present *or* the passed
    string as is otherwise.

    Parameters
    ----------
    text : str
        String to be examined. Since strings are immutable in Python, this
        string remains unmodified.
    suffix : str
        Suffix to remove from this string.

    Returns
    ----------
    str
        Resulting string as described above.
    '''

    # There exists a special case *NOT* present in remove_prefix(). If this
    # suffix is empty, "string[:-0]" is also incorrectly empty. Avoid returning
    # the empty string in this case by explicitly testing for emptiness.
    return text[:-len(suffix)] if suffix and is_suffix(text, suffix) else text


@type_check
def remove_suffix_with_prefix(text: str, suffix_prefix: str) -> str:
    '''
    Return the passed string without the first instance of the passed suffix
    prefix and all characters following this suffix prefix in this string if
    present *or* raise an exception otherwise.

    Parameters
    ----------
    text : str
        String to be examined. Since strings are immutable in Python, this
        string remains unmodified.
    suffix_prefix : str
        Non-empty substring of this string to begin removing characters at.

    Returns
    ----------
    str
        Resulting string as described above.

    Raises
    ----------
    BetseStrException
        If either:
        * This string does *not* contain this suffix prefix.
        * This suffix prefix is the empty string.
    '''

    # If this suffix prefix is the empty string, raise an exception.
    die_if_empty(text=text, exception_message='Suffix prefix empty.')

    # Attempt to return the desired string.
    try:
        return text[:text.index(suffix_prefix)]
    # If doing so fails with a builtin exception...
    except ValueError as exception:
        # ...failing to provide the contents of these arguments, wrap this
        # exception with a fine-grained exception providing these contents.
        if str(exception) == 'substring not found':
            raise BetseStrException(
                'Text "{}" contains no suffix prefix "{}".'.format(
                    text, suffix_prefix))

        # Else, re-raise this exception as is.
        raise

# ....................{ CASERS                             }....................
@type_check
def lowercase_char_first(text: str) -> str:
    '''
    Lowercase the first character of the passed string.
    '''

    return text[0].lower() + text[1:] if text else ''


@type_check
def uppercase_char_first(text: str) -> str:
    '''
    Uppercase the first character of the passed string.

    Whereas the related :meth:`str.capitalize` method both uppercases the first
    character of this string *and* lowercases all remaining characters, this
    function *only* uppercases the first character. All remaining characters
    remain unmodified.
    '''

    return text[0].upper() + text[1:] if text else ''

# ....................{ WRAPPERS                           }....................
def wrap_lines(lines: IterableTypes, **kwargs) -> str:
    '''
    Wrap the passed iterable of lines to the passed line width, prefixing each
    resulting wrapped line by the passed line prefix.

    See Also
    ----------
    wrap()
        For further details.
    '''

    return wrap(join(lines), **kwargs)


@type_check
def wrap(
    text: str,
    text_wrapper = textwrap,
    line_prefix: str = '',
    **kwargs
) -> str:
    '''
    Wrap the passed text to the passed line width, prefixing each resulting
    wrapped line by the passed line prefix.

    This function accepts the following keyword arguments in addition to those
    accepted by the :func:`textwrap.wrap` function:

    * ``text_wrapper`, the object on which to call the ``wrap()`` function or
      method. For safety, this defaults to the :mod:`textwrap` module.
    * ``line_prefix``, the substring prefixing each wrapped output line.

    See Also
    ----------
    https://docs.python.org/3/library/textwrap.html
        For further details on keyword arguments.
    '''

    assert hasattr(text_wrapper, 'wrap'), (
        'Object "{}" not a text wrapper '
        '(i.e., has no wrap() callable).'.format(text_wrapper))

    # wrap() function or method to be called.
    wrap_callable = getattr(text_wrapper, 'wrap')
    assert callable(wrap_callable), (
        'Text wrapper "{}" attribute "wrap" not callable.'.format(text_wrapper))

    # If passed a nonempty line prefix, add appropriate keyword arguments.
    if line_prefix:
        kwargs['initial_indent'] = line_prefix
        kwargs['subsequent_indent'] = line_prefix

    # For reliability, call the wrap() function of module "textwrap" rather than
    # the singleton object "text_wrapper". The former internally instantiates a
    # new instance of class "TextWrapper" and hence resets all wrapping
    # attributes to sensible defaults, whereas the latter reuses existing such
    # attributes -- which may no longer retain sensible defaults.
    return join_on_newline(wrap_callable(text, **kwargs))

# ....................{ WRAPPERS ~ un                      }....................
@type_check
def unwrap(text: str) -> str:
    '''
    Convert the passed possibly **multiline string** (i.e., string containing
    one or more newlines) to a **single-line string** (i.e., string containing
    no newlines).
    '''

    # Nice one, stdlib.
    return text.replace('\n', ' ')

# ....................{ (IN|DE)DENTERS                     }....................
def dedent(*texts) -> str:
    '''
    Remove all indentation shared in common by all lines of all passed strings.
    '''

    return textwrap.dedent(*texts)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Decorators marking tests and fixtures as being **parametrized** (i.e., accepting
two or more sets of parameters passed to these tests and fixtures).

Most of these decorators mark tests and fixtures with BETSE-specific keywords
inspected *only* by BETSE-specific py.test hooks defined by ``conftest``
plugins.
'''

# ....................{ IMPORTS                            }....................
import itertools, pytest
from betse_test.exceptions import BetseTestParamException
from betse.util.type.types import (
    type_check, CallableTypes, ContainerType, MappingType, SequenceTypes,)

# ....................{ PARAMS                             }....................
# Outer decorator accepting all parameters explicitly passed to this decorator.
@type_check
def parametrize_test_setwise(
    params: MappingType = None,
    fixtures: MappingType = None,
    ids: SequenceTypes = None,
) -> CallableTypes:
    '''
    Optionally directly parametrize all non-fixture parameters accepted by the
    decorated test callable with the first passed dictionary _and_ optionally
    indirectly parametrize all fixture parameters accepted by this callable with
    the second passed dictionary.

    Fixtures
    ----------
    This decorator should _not_ be directly applied to fixtures, for which the
    :func:`pytest.fixture` decorator should be applied instead as follows:

    * Pass that decorator:
      * The mandatory `params` keyword argument, whose value is the sequence of
        all sequences of parameters to be iteratively passed to that fixture.
      * The optional `ids` keyword argument, whose value is the sequence of all
        human-readable unique identifiers to be assigned to each such set of
        parameters (in the same order). While optional, this argument premains
        highly recommended.
    * Pass that fixture the `request` builtin fixture, whose `param` attribute
      supplies the set of parameters passed to the current fixture invocation.

    Fixture Parameters
    ----------
    For each fixture accepted by the decorated test callable, the
    :func:`pytest.fixture` decorator should be applied to that fixture as
    follows:

    * Pass that fixture the `request` builtin fixture, whose `param` attribute
      supplies the set of parameters passed to the current fixture invocation.

    Avoid passing that decorator either the `params` or `ids` keyword argument,
    whose values will be indirectly supplied by this decorator to that fixture.

    Parameters
    ----------
    params : optional[MappingType]
        Dictionary mapping from the names of non-fixture parameters accepted by
        this test to the sequences of all values of those parameters to be
        iteratively passed to this test (_in order_) such that:
        . The first element of each such sequence is the first value of the
          corresponding parameter to be passed to the first parametrization of
          this test.
        . The second element of the same sequence is the second value of this
          parameter to be passed to the second parametrization of this test.
        . And so on.
        Defaults to `None`, in which case this test is assumed to accept no
        non-fixture parameters.
    fixtures : optional[MappingType]
        Dictionary mapping from the names of fixture parameters accepted by this
        test to the sequences of all values of those parameters to be
        iteratively passed to this test (_in order_) such that the same
        interpretation as for `params` holds. Defaults to `None`, in which case
        this test is assumed to accept no fixture parameters.
    ids : optional[SequenceTypes]
        Sequence of all human-readable labels uniquely identifying each
        parametrization of this test (_in the same order_). Defaults to `None`,
        in which non-human-readable labels will be automatically synthesized
        from the values of the parameters comprising these parametrizations.

    Raises
    ----------
    BetseTestParamException
        If the `ids` argument identifies a different number of parametrizations
        than the `params` and `fixtures` arguments actually passed.

    Examples
    ----------
    >>> from pytest import fixture
    >>> from betse_test.util.mark.param import parametrize_test_setwise
    >>> @fixture
    ... def feathered_serpent(serpent_name: str) -> str:
    ...     if serpent_name == 'Hualpa'
    ...         return 'Amazonia'
    ...     elif serpent_name == 'Mujaji'
    ...         return 'Azania'
    >>> @parametrize_test_setwise(
    ...     params={
    ...         'western_dragon': ('Celedyr', 'Hestaby',),
    ...         'eastern_dragon': ('Masaru', 'Ryumyo',),
    ...     },
    ...     fixtures={
    ...         'feathered_serpent': ('Hualpa', 'Mujaji',),
    ...     },
    ...     ids=('bad-dragons', 'good-dragons',),
    ... )
    ... def test_great_dragons(
    ...     western_dragon: str, eastern_dragon: str, feathered_serpent: str):
    ...     assert western_dragon in ('Celedyr', 'Hestaby',)
    ...     assert eastern_dragon in ('Masaru', 'Ryumyo',)
    ...     assert feathered_serpent in ('Amazonia', 'Azania',)
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterables import zip_isometric

    # Default unpassed arguments to the empty dictionary for sanity. For unknown
    # (but presumably valid) reasons, these defaults *MUST* be established here
    # rather than in the inner closure defined below. Just. Because.
    if params   is None: params   = {}
    if fixtures is None: fixtures = {}

    # Inner closure decorating the actual test callable.
    def _parametrize_test_fixtures_inner(test_callable) -> CallableTypes:
        # Validate all non-fixture parametrizations to be sequences.
        for param_values in params.values():
            if not isinstance(param_values, SequenceTypes):
                raise BetseTestParamException(
                    'Parametrized test "{}" '
                    'non-fixture parametrization {} not a sequence.'.format(
                    test_callable.__name__, param_values))

        # Validate all fixture parametrizations to be sequences.
        for fixture_values in fixtures.values():
            if not isinstance(fixture_values, SequenceTypes):
                raise BetseTestParamException(
                    'Parametrized test "{}" '
                    'fixture parametrization {} not a sequence.'.format(
                    test_callable.__name__, fixture_values))

        # Tuple of the names of all fixture parameters (in arbitrary order) if
        # any or False otherwise. Attempting to pass an empty tuple as the
        # "indirect" keyword argument to the @pytest.mark.parametrize
        # decorator raises the following non-human-readable exception:
        #
        #     /usr/lib64/python3.4/site-packages/_pytest/python.py:1952: in pytest_generate_tests
        #         argnames = func_params[0]
        #     E   IndexError: tuple index out of range
        #
        # Likewise, attempting to pass None results in even more obscure
        # non-human-readable exceptions resembling:
        #
        #     /usr/lib64/python3.4/site-packages/_pytest/python.py:869: in setmulti
        #         valtype_for_arg = valtypes[arg]
        #     E   KeyError: '{param_name}'
        #
        # ...where "{param_name}" is the name of the first parameter listed in
        # the "param_fixture_names" tuple defined below.
        #
        # py.test, you are nice; but your failure to raise human-readable errors
        # is a deplorable waste worthy of unclean roadhouse outhouses.
        fixture_names = tuple(fixtures.keys()) or False

        # Tuple of the names of both non-fixture and fixture parameters (in a
        # predictable, but arbitrary, order).
        #
        # Note that the dict.keys() method returns instances of the "dict_keys"
        # class; unlike standard sequence classes (e.g., "list", "tuple"), this
        # class does *NOT* overload the "+" operator. Aggregating multiple
        # instances of this class requires leveraging the generic
        # itertools.chain() function, accepting arbitrary iterables. (Ugh.)
        param_fixture_names = tuple(itertools.chain(
            params.keys(), fixtures.keys()))
        # param_fixture_names = ','.join(itertools.chain(
        #     params.keys(), fixtures.keys()))

        # Tuple of n-tuples of all values for each parametrization of these
        # non-fixture and parameters (in this same predictable order). See
        # parametrize_test_setwise() for further discussion.
        param_fixture_values = tuple(zip_isometric(*itertools.chain(
            params.values(), fixtures.values())))

        # If the caller identified a different number of parametrizations than
        # were actually passed, raise an exception.
        if ids is not None and len(ids) != len(param_fixture_values):
            raise BetseTestParamException(
                'Parametrized test "{}" '
                'number of parametrization identifiers {} differs from '
                'number of parametrizations {}.'.format(
                 test_callable.__name__, len(ids), len(param_fixture_values)))
        # print('\n    test: {}; param_names: {!r}; values: {!r}; fixtures: {!r}; ids: {!r}'.format(test_callable.__name__, param_fixture_names, param_fixture_values, fixture_names, ids))

        # Defer to the canonical parametrize() decorator.
        return pytest.mark.parametrize(
            argnames=param_fixture_names,
            argvalues=param_fixture_values,
            indirect=fixture_names,
            ids=ids,
        )(test_callable)

    # Return this inner closure.
    return _parametrize_test_fixtures_inner

# ....................{ PARAMS ~ setwise                   }....................
# Outer decorator accepting all parameters explicitly passed to this decorator.

#FIXME: Actually implement. See the "Examples" docstring section for the
#intended usage of this decorator; basically, it provides an alternative means
#(and arguably equally intuitive) of parametrizing tests to the above
#parametrize_test_setwise() decorator.
#FIXME: The principal disadvantage of both this and the above
#parametrize_test_setwise() decorator is that neither supports the fine-grained
#specification of specific parametrizations to be either skipped or xfailed, as
#supported by the lower-level @pytest.mark.parametrize decorator. See the
#following official py.test documentation for a useful example:
#    http://doc.pytest.org/en/latest/skipping.html#skip-xfail-with-parametrize
#While the above parametrize_test_setwise() decorator *CANNOT* (by design) be
#refactored to support such fine-grained specification, this decorator should be
#able to do so. How? By defining two new "ParametrizationSkipIf" and
#"ParametrizationXFail" classes instantiated as follows:
#
#    @parametrize_test_setwise(
#        parametrizations={
#            'bad-dragons': {
#                'western_dragon':    'Celedyr',
#                'eastern_dragon':    'Masaru',
#                'feathered_serpent': 'Hualpa',
#            },
#            'good-dragons': ParametrizationXFail(
#                parametrizations={
#                    'western_dragon':    'Hestaby',
#                    'eastern_dragon':    'Ryumyo',
#                    'feathered_serpent': 'Mujaji',
#                },
#                reason="Good dragons raise unexpected breath attack.",
#            },
#        },
#        fixture_names={'feathered_serpent',},
#    )
#    def test_great_dragons(
#        western_dragon: str, eastern_dragon: str, feathered_serpent: str):
#        assert western_dragon in ('Celedyr', 'Hestaby',)
#        assert eastern_dragon in ('Masaru', 'Ryumyo',)
#        assert feathered_serpent in ('Amazonia', 'Azania',)
#
#This decorator would then need to explicitly differentiate such objects from
#standard dictionaries and behave as expected. Quite a bit of work, sadly, but
#certainly feasible.
#FIXME: Repair documentation.

@type_check
def parametrize_test_paramwise(
    fixture_names: ContainerType = None,
    parametrizations: MappingType = None,
) -> CallableTypes:
    '''
    Optionally directly parametrize all non-fixture parameters accepted by the
    decorated test callable with the first passed dictionary _and_ optionally
    indirectly parametrize all fixture parameters accepted by this callable with
    the second passed dictionary.

    Fixtures
    ----------
    This decorator should _not_ be directly applied to fixtures, for which the
    :func:`pytest.fixture` decorator should be applied instead as follows:

    * Pass that decorator:
      * The mandatory `params` keyword argument, whose value is the sequence of
        all sequences of parameters to be iteratively passed to that fixture.
      * The optional `ids` keyword argument, whose value is the sequence of all
        human-readable unique identifiers to be assigned to each such set of
        parameters (in the same order). While optional, this argument premains
        highly recommended.
    * Pass that fixture the `request` builtin fixture, whose `param` attribute
      supplies the set of parameters passed to the current fixture invocation.

    Fixture Parameters
    ----------
    For each fixture accepted by the decorated test callable, the
    :func:`pytest.fixture` decorator should be applied to that fixture as
    follows:

    * Pass that fixture the `request` builtin fixture, whose `param` attribute
      supplies the set of parameters passed to the current fixture invocation.

    Avoid passing that decorator either the `params` or `ids` keyword argument,
    whose values will be indirectly supplied by this decorator to that fixture.

    Parameters
    ----------
    params : optional[MappingType]
        Dictionary mapping from the names of non-fixture parameters accepted by
        this test to the sequences of all values of those parameters to be
        iteratively passed to this test (_in order_) such that:
        . The first element of each such sequence is the first value of the
          corresponding parameter to be passed to the first parametrization of
          this test.
        . The second element of the same sequence is the second value of this
          parameter to be passed to the second parametrization of this test.
        . And so on.
        Defaults to `None`, in which case this test is assumed to accept no
        non-fixture parameters.
    fixtures : optional[MappingType]
        Dictionary mapping from the names of fixture parameters accepted by this
        test to the sequences of all values of those parameters to be
        iteratively passed to this test (_in order_) such that the same
        interpretation as for `params` holds. Defaults to `None`, in which case
        this test is assumed to accept no fixture parameters.
    ids : optional[SequenceTypes]
        Sequence of all human-readable labels uniquely identifying each
        parametrization of this test (_in the same order_). Defaults to `None`,
        in which non-human-readable labels will be automatically synthesized
        from the values of the parameters comprising each such parametrization.

    Raises
    ----------
    BetseTestParamException
        If the `ids` argument identifies a different number of parametrizations
        than the `params` and `fixtures` arguments actually passed.

    Examples
    ----------
    >>> from pytest import fixture
    >>> from betse_test.util.mark.param import parametrize_test_setwise
    >>> @fixture
    ... def feathered_serpent(serpent_name: str) -> str:
    ...     if serpent_name == 'Hualpa'
    ...         return 'Amazonia'
    ...     elif serpent_name == 'Mujaji'
    ...         return 'Azania'
    >>> @parametrize_test_setwise(
    ...     parametrizations={
    ...         'bad-dragons': {
    ...             'western_dragon':    'Celedyr',
    ...             'eastern_dragon':    'Masaru',
    ...             'feathered_serpent': 'Hualpa',
    ...         },
    ...         'good-dragons': {
    ...             'western_dragon':    'Hestaby',
    ...             'eastern_dragon':    'Ryumyo',
    ...             'feathered_serpent': 'Mujaji',
    ...         },
    ...     },
    ...     fixture_names={'feathered_serpent',},
    ... )
    ... def test_great_dragons(
    ...     western_dragon: str, eastern_dragon: str, feathered_serpent: str):
    ...     assert western_dragon in ('Celedyr', 'Hestaby',)
    ...     assert eastern_dragon in ('Masaru', 'Ryumyo',)
    ...     assert feathered_serpent in ('Amazonia', 'Azania',)
    '''

    raise ValueError('This decorator currently unimplemented.')

# ....................{ PARAMS ~ serial                    }....................
serialize_parametrized_test = pytest.mark.serialize_parametrized_test
'''
Mark the decorated parametrized test as **serial** (i.e., parametrized such that
the success of each subsequent parameter set depends on the success of all
previous parameter sets for the current test requesting this decoration).

If this test is _not_ parametrized, an exception is raised. Else, each parameter
set passed to this test is tested serially. On the first failure of this test
passed a parameter set, each subsequent call to this test passed a subsequent
parameter set will be automatically marked as an `XFAIL` and hence fail
_without_ being run.

Implementation
----------
The majority of the black magic required by this decoration is implemented as
low-level `py.test` hooks in the top-level :mod:`betse_test.conftest` plugin. To
preserve state between parametrized calls to the same test, these hooks
dynamically add the following BETSE-specific attributes to this test's
underlying function or method object:

* `_betse_first_failing_param_id`, the unique identifier of the first set of
  parameter values passed to this test raising an exception for the current test
  session if any _or_ `None` otherwise (i.e., if this test has yet to raise an
  exception for any parameters).
'''


def serialize_parametrized_fixture(fixture):
    '''
    Decorator marking the decorated parametrized fixture as **serial** (i.e.,
    parametrized such that the success of each subsequent parameter set depends
    on the success of all previous parameter sets for the current test
    requesting this decoration).

    If this fixture is _not_ parametrized, an exception is raised.

    This decoration is automatically propagated to all tests requesting this
    fixture. Decorating a fixture with this decorator is functionally
    equivalent to decorating all tests requesting this fixture by the
    `@serialize_parametrized_test` decorator.

    Implementation
    ----------
    Technically, `py.test` currently ignores marks decorating fixtures. To
    circumvent this omission, this decorator adds the following BETSE-specific
    attribute to this fixture's underlying function or method object:

    * `_betse_is_fixture_parametrized_serially`, `True` only if this fixture is
      parametrized serially. If this fixture is _not_ parametrized serially,
      this attribute is usually undefined.

    See Also
    ----------
    :func:`serialize_parametrized_test`
        Further details on serial parametrization.
    '''

    # Mark this fixture as serially parametrized. Since fixtures do *NOT*
    # support marks, a BETSE-specific attribute is added to this fixturegg
    fixture._betse_is_fixture_parametrized_serially = True
    return fixture



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`setuptools`-based makefile instrumenting all high-level administration
tasks (e.g., installation, freezing, test running) for this application.
'''

#FIXME: Consider wrapping on Windows with "pynsist", a framework for generating
#Windows installers bundling Python applications complete with a Python
#interpreter and requisite packages.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from packages guaranteed to exist at the start of
# installation. This includes all standard Python and application packages but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#
# Technically, this script may import from all subpackages and submodules of the
# this application's eponymous package. By Python mandate, the first element of
# the "sys.path" list is guaranteed to be the directory containing this script.
# Python necessarily searches this directory for imports from the local version
# of this application *BEFORE* any other directories (including system
# directories containing older versions of this application). To quote:
#
#     "As initialized upon program startup, the first item of this list,
#      path[0], is the directory containing the script that was used to invoke
#      the Python interpreter."
#
# See also: https://stackoverflow.com/a/10097543/2809027
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import setuptools
from betse import metadata
from betse.lib import libs
from betse_setup import build, freeze, symlink, test, buputil

# ....................{ METADATA                           }....................
# PyPI-specific metadata declared here rather than in the "betse.metadata"
# submodule, reducing space and time complexity during application startup. This
# metadata is relevant only to setuptools and hence irrelevant to the main
# codebase.

_PYTHON_VERSION_MINOR_MAX = 7
'''
Maximum minor stable version of this major version of Python currently released
(e.g., ``5`` if Python 3.5 is the most recent stable version of Python 3.x).
'''


_DESCRIPTION = None
'''
Human-readable multiline description of this application in reStructuredText
(reST) format.

To minimize synchronization woes, this description is identical to the contents
of the :doc:`/README.rst` file. When submitting this application package to
PyPI, this description is used verbatim as this package's front matter.
'''


# ....................{ METADATA ~ seo                     }....................
_KEYWORDS = ['biology', 'multiphysics', 'science', 'simulator',]
'''
List of all lowercase alphabetic keywords synopsising this application.

These keywords may be arbitrarily selected so as to pretend to improve search
engine optimization (SEO). In actuality, they do absolutely nothing.
'''


# To minimize desynchronization woes, all
# "Programming Language :: Python :: "-prefixed strings are dynamically appended
# to this list by the init() function below.
_CLASSIFIERS = [
    #FIXME: Replace with the following after the "1.0.0" release:
    #    'Development Status :: 5 - Production/Stable',

    # PyPI-specific version type. The number specified here is a magic constant
    # with no relation to this application's version numbering scheme. *sigh*
    'Development Status :: 4 - Beta',

    # Sublist of all supported platform-specific CLI and GUI components.
    'Environment :: Console',
    'Environment :: MacOS X',
    'Environment :: Win32 (MS Windows)',
    'Environment :: X11 Applications',

    # Miscellaneous metadata.
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
]
'''
List of all PyPI-specific trove classifier strings synopsizing this application.

Each such string *must* be contain either two or three `` :: `` substrings
delimiting human-readable capitalized English words formally recognized by the
:mod:`distutils`-specific ``register`` command.

See Also
----------
https://pypi.python.org/pypi?%3Aaction=list_classifiers
    Plaintext list of all trove classifier strings recognized by PyPI.
'''

# ....................{ INITIALIZERS                       }....................
def _init() -> None:
    '''
    Finalize the definition of all globals declared by this script.
    '''

    # Global variables assigned to below.
    global _DESCRIPTION

    # Major version of Python required by this application.
    PYTHON_VERSION_MAJOR = metadata.PYTHON_VERSION_MIN_PARTS[0]

    # Relative path of this application's front-facing documentation in
    # reStructuredText format, required by PyPI.
    DESCRIPTION_FILENAME = 'README.rst'

    # For each minor version of Python 3.x supported by this application,
    # formally classify this version as such.
    for python_version_minor in range(
        metadata.PYTHON_VERSION_MIN_PARTS[1], _PYTHON_VERSION_MINOR_MAX + 1):
        _CLASSIFIERS.append(
            'Programming Language :: Python :: {}.{}'.format(
                PYTHON_VERSION_MAJOR, python_version_minor,))
    # print('classifiers: {}'.format(_CLASSIFIERS))

    # Description read from this description file.
    try:
        _DESCRIPTION = buputil.get_chars(DESCRIPTION_FILENAME)
        # print('description: {}'.format(_DESCRIPTION))
    # If this file is *NOT* readable, print a non-fatal warning and reduce this
    # description to the empty string. While unfortunate, this description is
    # *NOT* required for most operations and hence mostly ignorable.
    except Exception as exception:
        _DESCRIPTION = ''
        buputil.output_warning(
            'Description file "{}" not found or not readable:\n{}'.format(
                DESCRIPTION_FILENAME, exception))

# ....................{ INITIALIZERS ~ main                }....................
# Finalize the definition of all globals declared by this module.
_init()

#FIXME: O.K.; it would appear that setuptools >= 38.0.0 now requires iterables
#to strictly be ordered sequences, whereas before it accepted unordered
#sequences. So, setuptools broke backwards compatibility yet again. Fortunately,
#the fix should be (mostly) trivial: just compact the
#"metadeps.RUNTIME_MANDATORY" dictionary passed to the "install_requires"
#parameter into a tuple of concatenated strings instead. We'll also need to do
#so for any other parameters accepting similar iterables. *sigh*

# print('mandatory runtime dependencies: {}'.format(metadeps.RUNTIME_MANDATORY))

# ....................{ OPTIONS                            }....................
# Setuptools-specific options. Keywords not explicitly recognized by either
# setuptools or distutils must be added to the above dictionary instead.
_setup_options = {
    # ..................{ CORE                               }..................
    # Self-explanatory metadata.
    'name':             metadata.PACKAGE_NAME,
    'version':          metadata.VERSION,
    'author':           metadata.AUTHORS,
    'author_email':     metadata.AUTHOR_EMAIL,
    'maintainer':       metadata.AUTHORS,
    'maintainer_email': metadata.AUTHOR_EMAIL,
    'description':      metadata.SYNOPSIS,
    'long_description': _DESCRIPTION,
    'url':              metadata.URL_HOMEPAGE,
    'download_url':     metadata.URL_DOWNLOAD,

    # ..................{ PYPI                               }..................
    # PyPi-specific metadata.
    'classifiers': _CLASSIFIERS,
    'keywords': _KEYWORDS,
    'license': metadata.LICENSE,

    # ..................{ DEPENDENCIES                       }..................
    # Mandatory runtime dependencies.
    'install_requires': libs.get_runtime_mandatory_tuple(),

    # Optional nuntime dependencies. Whereas mandatory dependencies are defined
    # as sequences, optional dependencies are defined as a dictionary mapping
    # from an arbitrary alphanumeric word to a sequence containing one or more
    # such dependencies. Such dependencies are then installable via "pip" by
    # suffixing the name of this project by the "["- and "]"-delimited key
    # defined below whose value lists the dependencies to be installed (e.g.,
    # "sudo pip3 install betse[all]", installing both the application and all
    # mandatory and optional dependencies required by the application).
    'extras_require': {
        'all': libs.get_runtime_optional_tuple(),
    },

    # Mandatory testing dependencies.
    'tests_require': libs.get_testing_mandatory_tuple(),

    # ..................{ PACKAGES                           }..................
    # List of all Python packages (i.e., directories containing zero or more
    # Python modules) to be installed. Currently, this includes the "betse"
    # package and all subpackages of this package excluding:
    #
    # * The top-level test package and all subpackages of this package, test
    #   functionality *NOT* intended to be installed with this application.
    # * The top-level setup package and all subpackages of this package,
    #   setuptools functionality required only for application installation.
    # * "build", caching both setuptools metadata and a complete copy of this
    #   package, required only by a prior application installation.
    # * "freeze", providing PyInstaller-specific functionality required only for
    #   application freezing (i.e., conversion into an executable binary).
    'packages': setuptools.find_packages(exclude=(
        metadata.PACKAGE_NAME + '_test',
        metadata.PACKAGE_NAME + '_test.*',
        metadata.PACKAGE_NAME + '_setup',
        metadata.PACKAGE_NAME + '_setup.*',
        'build',
        'freeze',
    )),

    # ..................{ PATHS                              }..................
    # Cross-platform script wrappers dynamically created at installation time.
    'entry_points': {
        # CLI-specific scripts.
        'console_scripts': (
            '{} = {}.__main__:main'.format(
                metadata.SCRIPT_BASENAME, metadata.PACKAGE_NAME),
        ),
    },

    #FIXME; This isn't quite true. Undesirable files are excludable in this
    #file via the whitelist approach of "package_data" and/or blacklist approach
    #of "exclude_package_data". This functionality may or may not require the
    #external "setuptools-git" plugin, but is certainly feasible. See also the
    #comprehensive documentation at:
    #https://pythonhosted.org/setuptools/setuptools.html#including-data-files
    #FIXME: O.K.; if "'include_package_data': True" ends up not working for us,
    #contemplate:
    #
    #    'package_data': {
    #        # Include all files in all packages matching the following.
    #        '': ['*.txt', '*.xml', '*.special', '*.huh'],
    #    },

    # Install all data files (i.e., non-Python files) embedded in the Python
    # package tree for this application.
    #
    # Unlike Python packages, undesirable data files are excludable from
    # installation *ONLY* via the external "MANIFEST.in" file. This is
    # terrible, of course. (Did you expect otherwise?)
    #
    # Data files are *NOT* Python modules and hence should *NOT* be embedded in
    # the Python package tree. Sadly, the "data_files" key supported by
    # setuptools for this purpose is *NOT* cross-platform-portable and is thus
    # inherently broken. Why? Because this key either requires usage of absolute
    # paths *OR* relative paths relative to absolute paths defined by
    # "setup.cfg"; in either case, these paths are absolute. While the current
    # platform could be detected and the corresponding absolute path embedded in
    # 'data_files', that implementation would be inherently fragile. (That's
    # bad.) In lieu of sane setuptools support, we defer to the methodology
    # employed by everyone. Setuptools, your death is coming.
    'include_package_data': True,

    # Install to an uncompressed directory rather than a compressed archive.
    #
    # While nothing technically precludes the latter, doing so substantially
    # complicates runtime access of data files compressed into this archive
    # (e.g., with the pkg_resources.resource_filename() function). How so? By
    # decompressing this archive's contents into a temporary directory on
    # program startup and removing these contents on program shutdown. Since
    # there exists no guarantee this removal will actually be performed (e.g.,
    # due to preemptive SIGKILLs), compressed archives are inherently fragile.
    'zip_safe': False,

    # ..................{ COMMANDS                           }..................
    # Set of all custom setuptools subcommands specific to this makefile (e.g.,
    # "sudo python3 setup.py symlink"), defaulting to the empty set. Each
    # subsequent call to the add_setup_commands() function iteratively performed
    # below adds one or more such subcommands to this set.
    'cmdclass': {},
}
'''
Dictionary passed to the subsequent call to the :func:`setup` function.

This dictionary signifies the set of all application-specific :mod:`setuptools`
options. Submodules of the top-level :mod:`betse_setup` package subsequently
customize these options (e.g., by defining custom commands).
'''
# print('extras: {}'.format(setup_options['extras_require']))


_setup_options_custom = {
    # While currently empty, it's likely we'll want this again... someday.
}
'''
Non-setuptools-specific metadata, used to inform custom subcommands (e.g.,
``freeze_file``) of other metadata *not* already declared by the
:data:`setup_options` dictionary.

Setuptools raises fatal exceptions when the :data:`setup_options` dictionary
contains unrecognized keys. For safety, these keys are added to this dictionary
instead.
'''

# ....................{ COMMANDS                           }....................
# Define all application-specific setuptools commands.
for setup_module in (build, freeze, symlink, test):
    setup_module.add_setup_commands(_setup_options_custom, _setup_options)

# ....................{ SETUP                              }....................
setuptools.setup(**_setup_options)



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Tissue profile pickers** (i.e., objects assigning a subset of all cells
matching various criteria to the corresponding tissue profile) classes.
'''

# ....................{ IMPORTS                            }....................
from abc import ABCMeta, abstractmethod
from betse.exceptions import BetseSimConfException
from betse.lib.numpy import nparray
from betse.science.math import toolbox
from betse.util.io.log import logs
from betse.util.type.types import type_check, NumericSimpleTypes, SequenceTypes
import numpy as np

# ....................{ SUPERCLASS                         }....................
class TissuePickerABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **tissue profile picker** (i.e., object assigning
    a subset of all cells matching some criteria to the corresponding tissue
    profile) subclasses.

    Typical criteria for matching cells includes explicit cell indexing,
    randomized cell selection, and image-defined spatial cell location.
    '''

    # ..................{ PICKERS                            }..................
    @abstractmethod
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:
        '''
        One-dimensional Numpy array of the indices of all cells in the passed
        cell cluster selected by this tissue picker, ignoring extracellular
        spaces.

        Parameters
        ----------
        cells : Cells
            Current cell cluster.
        p : Parameters
            Current simulation configuration.

        Returns
        ----------
        SequenceTypes
            One-dimensional Numpy array of the indices of all such cells.
        '''

        pass


    @type_check
    def pick_cells_and_mems(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> tuple:
        '''
        2-tuple ``(cells_index, mems_index)`` of one-dimensional Numpy arrays of
        the indices of both all cells *and* cell membranes in the passed cell
        cluster selected by this tissue picker, ignoring extracellular spaces.

        By default, this method returns the array returned by the subclass
        implementation of the abstract :meth:`pick_cells` method, mapped from
        cell to cell membrane indices.

        Parameters
        ----------
        cells : Cells
            Current cell cluster.
        p : Parameters
            Current simulation configuration.

        Returns
        ----------
        (ndarray, ndarray)
            2-tuple ``(cells_index, mems_index)``, where:
            * ``cells_index`` is the one-dimensional Numpy array of the indices
              of all subclass-selected cells.
            * ``mems_index`` is the one-dimensional Numpy array of the indices
              of all subclass-selected cell membranes.
        '''

        # One-dimensional Numpy array of the indices of subclass-selected cells.
        cells_index = self.pick_cells(cells, p)

        #FIXME: Ideally, this array would be trivially defined as follows:
        #    mems_index = cells.cell_to_mems[cells_index].flatten()
        #Unfortunately, the two-dimensional Numpy array "cells.cell_to_mems"
        #actually appears to be a one-dimensional array of lists -- which is a
        #bit bizarro-world. To compaund matters, this array's "dtype" is
        #"object" (presumably, because it contains Python lists) rather than the
        #"dtype" of "int" that one might expect. Since this is the case, it's
        #infeasible to even coerce the temporary one-dimensional Numpy array
        #"cells.cell_to_mems[cells_index]" into a two-dimensional array of ints
        #by calling ndarray.astype(int). Frankly, it's all a bit beyond me;
        #until this core issue is resolved, however, the current inelegant and
        #inefficient approach remains.

        # One-dimensional Numpy array of the indices of subclass-selected cell
        # membranes mapped from the array of cell indices.
        # logs.log_debug('cells_index: %r', cells_index)
        mems_index = cells.cell_to_mems[cells_index]
        mems_index, _, _ = toolbox.flatten(mems_index)
        mems_index = nparray.from_iterable(mems_index)

        # Return these arrays.
        return cells_index, mems_index

# ....................{ SUBCLASSES                         }....................
class TissuePickerAll(TissuePickerABC):
    '''
    All-inclusive tissue picker, unconditionally matching *all* cells in the
    current cell cluster.
    '''

    # ..................{ PICKERS                            }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        return cells.cell_i


class TissuePickerIndices(TissuePickerABC):
    '''
    Cell indices-based tissue picker, matching all cells in the current cell
    cluster whose indices are defined by a given sequence.

    Attributes
    ----------
    cells_index : SequenceTypes
        Sequence of the indices of all cells to match.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, cells_index: SequenceTypes) -> None:
        '''
        Initialize this tissue picker.

        Parameters
        ----------
        cells_index : SequenceTypes
            Sequence of the indices of all cells to match.
        '''

        self.cells_index = cells_index

    # ..................{ PICKERS                            }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        return self.cells_index


class TissuePickerPercent(TissuePickerABC):
    '''
    Randomized cell picker, randomly matching a given percentage of all cells in
    the current cell cluster.

    Attributes
    ----------
    cells_percent : NumericSimpleTypes
        **Percentage** (i.e., number in the range ``[0.0, 100.0]``) of the total
        cell population to randomly match.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, cells_percent: NumericSimpleTypes) -> None:
        '''
        Initialize this tissue picker.

        Parameters
        ----------
        cells_percent : NumericSimpleTypes
            **Percentage** (i.e., number in the range ``[0.0, 100.0]``) of the
            total cell population to randomly match.
        '''

        # If this is not a valid percentage, raise an exception. This is
        # important enough to always test rather than defer to assertions.
        if not 0.0 <= cells_percent <= 100.0:
            raise BetseSimConfException(
                '{} not in the range [0.0, 100.0].'.format(cells_percent))

        # Classify this parameter.
        self.cells_percent = cells_percent

    # ..................{ PICKERS                            }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        # Total number of cells in this cluster.
        data_length = len(cells.cell_i)

        # Total number of cells to randomly select from this cluster.
        data_fraction = int((self.cells_percent/100)*data_length)

        cell_i_copy = cells.cell_i[:]
        np.random.shuffle(cell_i_copy)

        # For simplicity, non-randomly select the indices of the first
        # "data_fraction"-th cells in this cluster. While technically
        # non-random, this simplistic approach currently suffices.
        return [cell_i_copy[x] for x in range(0, data_fraction)]



#!/usr/bin/env python3
# Copyright 2014-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Post-simulation single-cell plot pipelines** (i.e., pipelines plotting
simulated data of a single cell of the cell cluster).
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.exceptions import BetseSimConfException
from betse.science.config.export.visual.confvisabc import (
    SimConfVisualCellListItem)
from betse.science.phase.pipe.piperun import piperunner
from betse.science.phase.require import phasereqs
from betse.science.visual.plot import plotutil
from betse.science.visual.plot.pipe.plotpipeabc import PlotPipeABC
from betse.util.type.types import IterableTypes
from matplotlib import pyplot as pyplot

# ....................{ SUBCLASSES                         }....................
#FIXME: Rename to "SimPipeExportPlotsCell" for disambiguity.
class PlotCellPipe(PlotPipeABC):
    '''
    **Post-simulation single-cell plot pipeline** (i.e., object iteratively
    displaying and/or saving all plots specific to a single cell of the cell
    cluster, produced after initialization and simulation as specified by the
    current simulation configuration).
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, label_singular='single-cell plot', **kwargs)

        # If this index is not that of an actual cell, raise an exception.
        if self._phase.p.plot_cell not in self._phase.cells.cell_i:
            raise BetseSimConfException(
                'Plot cell index {} invalid '
                '(i.e., not in range [{}, {}]).'.format(
                    self._phase.p.plot_cell,
                    self._phase.cells.cell_i[0],
                    self._phase.cells.cell_i[-1],
                ))

    # ..................{ SUPERCLASS                         }..................
    @property
    def _runners_conf(self) -> IterableTypes:
        return self._phase.p.plot.plots_cell_after_sim

    # ..................{ EXPORTERS ~ cell : current         }..................
    @piperunner(
        categories=('Current Density', 'Transmembrane',),
        requirements=phasereqs.ELECTRIC_CURRENT_MEMBRANE,
    )
    def export_currents_membrane(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot all transmembrane current densities for the single cell indexed by
        the current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        pyplot.figure()
        axI = pyplot.subplot(111)

        if self._phase.p.is_ecm:
            # Total cell current storage vector.
            Imem = []

            # Indices of all membranes for this cell.
            mems_for_plotcell = self._phase.cells.cell_to_mems[
                self._phase.p.plot_cell]

            for t in range(len(self._phase.sim.time)):
                memArray = self._phase.sim.I_mem_time[t]

                # X and Y components of the net current at each membrane.
                Ixo = (
                    memArray[mems_for_plotcell] *
                    self._phase.cells.mem_vects_flat[mems_for_plotcell,2] *
                    self._phase.cells.mem_sa[mems_for_plotcell]
                )
                Iyo = (
                    memArray[mems_for_plotcell] *
                    self._phase.cells.mem_vects_flat[mems_for_plotcell,3] *
                    self._phase.cells.mem_sa[mems_for_plotcell]
                )

                # X and Y components of the net current over all membranes,
                # implicitly accounting for in-out directionality.
                Ix = np.sum(Ixo)
                Iy = np.sum(Iyo)

                # Current density, obtained by dividing the total magnitude of
                # the net current by this cell's surface area.
                Io = np.sqrt(Ix**2 + Iy**2) / self._phase.cells.cell_sa[
                    self._phase.p.plot_cell]
                Imem.append(100*Io)
        else:
            Imem = [
                100*memArray[self._phase.p.plot_cell]
                for memArray in self._phase.sim.I_mem_time
            ]

        axI.plot(self._phase.sim.time, Imem)
        axI.set_xlabel('Time [s]')
        axI.set_ylabel('Current density [uA/cm2]')
        axI.set_title(
            'Transmembrane current density for cell {}'.format(
                self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='Imem_time')

    # ..................{ EXPORTERS ~ cell : deform          }..................
    @piperunner(
        categories=('Deformation', 'Total',),
        requirements=phasereqs.DEFORM,
    )
    def export_deform_total(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot all **total cellular displacements** (i.e., summations of all
        cellular deformations due to galvanotropic and osmotic pressure body
        forces) for the single cell indexed by the current simulation
        configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        # Extract time-series deformation data for the plot cell.
        dx = np.asarray([
            arr[self._phase.p.plot_cell]
            for arr in self._phase.sim.dx_cell_time])
        dy = np.asarray([
            arr[self._phase.p.plot_cell]
            for arr in self._phase.sim.dy_cell_time])

        # Get the total magnitude.
        disp = np.sqrt(dx**2 + dy**2)

        pyplot.figure()
        axD = pyplot.subplot(111)
        axD.plot(self._phase.sim.time, self._phase.p.um*disp)
        axD.set_xlabel('Time [s]')
        axD.set_ylabel('Displacement [um]')
        axD.set_title('Displacement of cell {}'.format(self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='Displacement_time')

    # ..................{ EXPORTERS ~ cell : ion             }..................
    @piperunner(
        categories=('Ion Concentration', 'Calcium'),
        requirements=phasereqs.ION_CALCIUM,
    )
    def export_ion_calcium(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot all calcium (i.e., Ca2+) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        figA, axA = plotutil.plotSingleCellCData(
            self._phase.sim.cc_time,
            self._phase.sim.time,
            self._phase.sim.iCa,
            self._phase.p.plot_cell,
            fig=None,
            ax=None,
            lncolor='g',
            ionname='Ca2+ cell',
        )
        axA.set_title('Cytosolic Ca2+ in cell {}'.format(
            self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='cytosol_Ca_time')


    @piperunner(
        categories=('Ion Concentration', 'M anion'),
        requirements=phasereqs.ION_M_ANION,
    )
    def export_ion_m_anion(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot all M anion (i.e., M-) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        figConcsM, axConcsM = plotutil.plotSingleCellCData(
            self._phase.sim.cc_time,
            self._phase.sim.time,
            self._phase.sim.iM,
            self._phase.p.plot_cell,
            fig=None, ax=None,
            lncolor='r',
            ionname='M-',
        )
        axConcsM.set_title(
            'M Anion concentration in cell {}'.format(self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='concM_time')


    @piperunner(
        categories=('Ion Concentration', 'Potassium'),
        requirements=phasereqs.ION_POTASSIUM,
    )
    def export_ion_potassium(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot all potassium (i.e., K+) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        figConcsK, axConcsK = plotutil.plotSingleCellCData(
            self._phase.sim.cc_time,
            self._phase.sim.time,
            self._phase.sim.iK,
            self._phase.p.plot_cell,
            fig=None, ax=None,
            lncolor='b',
            ionname='K+',
        )
        axConcsK.set_title(
            'Potassium concentration in cell {}'.format(
                self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='concK_time')


    @piperunner(
        categories=('Ion Concentration', 'Sodium'),
        requirements=phasereqs.ION_SODIUM,
    )
    def export_ion_sodium(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot all sodium (i.e., Na+) ion concentrations for the single cell
        indexed by the current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        # Plot cell sodium concentration versus time.
        figConcsNa, axConcsNa = plotutil.plotSingleCellCData(
            self._phase.sim.cc_time,
            self._phase.sim.time,
            self._phase.sim.iNa,
            self._phase.p.plot_cell,
            fig=None,
            ax=None,
            lncolor='g',
            ionname='Na+',
        )
        axConcsNa.set_title(
            'Sodium concentration in cell {}'.format(self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='concNa_time')

    # ..................{ EXPORTERS ~ pressure               }..................
    @piperunner(
        categories=('Pressure', 'Osmotic',),
        requirements=phasereqs.PRESSURE_OSMOTIC,
    )
    def export_pressure_osmotic(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot the osmotic cellular pressure for the single cell indexed by the
        current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        p_osmo = tuple(
            arr[self._phase.p.plot_cell]
            for arr in self._phase.sim.osmo_P_delta_time)

        pyplot.figure()
        axOP = pyplot.subplot(111)
        axOP.plot(self._phase.sim.time, p_osmo)
        axOP.set_xlabel('Time [s]')
        axOP.set_ylabel('Osmotic Pressure [Pa]')
        axOP.set_title(
            'Osmotic pressure in cell {}'.format(self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='OsmoticP_time')


    @piperunner(
        categories=('Pressure', 'Total',),
        requirements=phasereqs.PRESSURE_TOTAL,
    )
    def export_pressure_total(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot the **total cellular pressure** (i.e., summation of the cellular
        mechanical and osmotic pressure) for the single cell indexed by the
        current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        p_hydro = tuple(
            arr[self._phase.p.plot_cell]
            for arr in self._phase.sim.P_cells_time)

        pyplot.figure()
        axOP = pyplot.subplot(111)
        axOP.plot(self._phase.sim.time, p_hydro)
        axOP.set_xlabel('Time [s]')
        axOP.set_ylabel('Total Pressure [Pa]')
        axOP.set_title(
            'Total pressure in cell {}'.format(self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='P_time')

    # ..................{ EXPORTERS ~ cell : pump            }..................
    #FIXME: Actually plot the average of these rates. Currently, this method
    #only plots rates for a single arbitrarily selected membrane of this cell.
    @piperunner(
        categories=('Pump Rate', 'Na-K-ATPase',),
        requirements=phasereqs.PUMP_NAKATPASE,
    )
    def export_pump_nakatpase(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot the averages of the Na-K-ATPase membrane pump rates for the single
        cell indexed by the current simulation configuration over all time
        steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        pyplot.figure()
        axNaK = pyplot.subplot()

        if self._phase.p.is_ecm:
            #FIXME: Generalize to average across all membranes of this cell.
            pump_array_index = self._phase.cells.cell_to_mems[
                self._phase.p.plot_cell][0]
        else:
            pump_array_index = self._phase.p.plot_cell

        pump_rate = tuple(
            pump_array[pump_array_index]
            for pump_array in self._phase.sim.rate_NaKATP_time)

        axNaK.plot(self._phase.sim.time, pump_rate)
        axNaK.set_xlabel('Time [s]')
        axNaK.set_ylabel('Pumping Rate of Na+ Out of Cell [mol/(m2 s)]')
        axNaK.set_title(
            'Rate of NaK-ATPase pump in cell: {}'.format(
                self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='NaKATPase_rate')

    # ..................{ EXPORTERS ~ cell : voltage         }..................
    #FIXME: Actually plot the average of these Vmems. Currently, this method
    #only plots Vmems for a single arbitrarily selected membrane of this cell.
    @piperunner(categories=('Voltage', 'Transmembrane', 'Average',))
    def export_voltage_membrane(self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot the averages of all transmembrane voltages for the single cell
        indexed by the current simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        #FIXME: Generalize to average across all membranes of this cell.
        mem_i = self._phase.cells.cell_to_mems[self._phase.p.plot_cell][0]

        # Plot single cell Vmem vs time.
        figVt, axVt = plotutil.plotSingleCellVData(
            self._phase.sim,
            mem_i,
            self._phase.p,
            fig=None, ax=None,
            lncolor='k',
        )
        axVt.set_title(
            'Voltage (Vmem) in cell {}'.format(self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='Vmem_time')


    #FIXME: Actually plot the average of these Vmems. Currently, this method
    #only plots Vmems for a single arbitrarily selected membrane of this cell.
    @piperunner(categories=(
        'Voltage', 'Transmembrane', 'Fast Fourier Transform (FFT)',))
    def export_voltage_membrane_fft(
        self, conf: SimConfVisualCellListItem) -> None:
        '''
        Plot the fast Fourier transform (FFT) of the averages of all
        transmembrane voltages for the single cell indexed by the current
        simulation configuration over all sampled time steps.
        '''

        # Prepare to export the current plot.
        self._export_prep()

        #FIXME: Generalize to average across all membranes of this cell.
        mem_i = self._phase.cells.cell_to_mems[self._phase.p.plot_cell][0]

        figFFT, axFFT = plotutil.plotFFT(
            self._phase.sim.time, self._phase.sim.vm_time, mem_i, lab="Power")
        axFFT.set_title(
            'Fourier transform of Vmem in cell {}'.format(
                self._phase.p.plot_cell))

        # Export this plot to disk and/or display.
        self._export(basename='Vmem_FFT_time')
